@@ -1,11 +1,11 @@
 
 /*
- *           PVM 3.0:  Parallel Virtual Machine System 3.0
+ *           PVM 3.2:  Parallel Virtual Machine System 3.2
  *               University of Tennessee, Knoxville TN.
  *           Oak Ridge National Laboratory, Oak Ridge TN.
  *                   Emory University, Atlanta GA.
  *      Authors:  A. L. Beguelin, J. J. Dongarra, G. A. Geist,
- *          R. J. Manchek, B. K. Moore, and V. S. Sunderam
+ *    W. C. Jiang, R. J. Manchek, B. K. Moore, and V. S. Sunderam
  *                   (C) 1992 All Rights Reserved
  *
  *                              NOTICE
@@ -22,7 +22,7 @@
  * purpose.  This software is provided ``as is'' without express or
  * implied warranty.
  *
- * PVM 3.0 was funded in part by the U.S. Department of Energy, the
+ * PVM 3.2 was funded in part by the U.S. Department of Energy, the
  * National Science Foundation and the State of Tennessee.
  */
 
@@ -31,7 +31,10 @@
  *
  *	Entry points for messages from network.
  *
-$Log$
+$Log: ddpro.c,v $
+ * Revision 1.1  1993/08/30  23:26:47  manchek
+ * Initial revision
+ *
  */
 
 
@@ -58,14 +61,22 @@ $Log$
 #include "waitc.h"
 #include "task.h"
 #include "listmac.h"
+#if defined(IMA_PGON) || defined(IMA_I860) || defined(IMA_CM5)
+#include "pvmmimd.h"
+#endif
+#include "bfunc.h"
 
 
 extern void pvmbailout();
 extern char *inadport_hex();
+char **pathsep();
 
 extern int debugmask;				/* from pvmd.c */
+extern char **epaths;				/* from pvmd.c */
+extern char *debugger;				/* from pvmd.c */
 extern struct htab *hosts;			/* from pvmd.c */
 extern struct task *locltasks;		/* from task.c */
+extern char *myarchname;			/* from pvmd.c */
 extern int myhostpart;				/* from pvmd.c */
 extern int mytid;					/* from pvmd.c */
 extern int myunixpid;				/* from pvmd.c */
@@ -75,6 +86,7 @@ extern int pprime;					/* from pvmd.c */
 extern int pprwid;					/* from pvmd.c */
 extern int runstate;				/* from pvmd.c */
 extern int tidhmask;				/* from pvmd.c */
+extern int tidlmask;				/* from pvmd.c */
 extern struct waitc *waitlist;		/* from waitc.c */
 
 
@@ -83,12 +95,11 @@ extern struct waitc *waitlist;		/* from waitc.c */
  **           **
  ***************/
 
-static char rcsid[] = "$Id$";
+static char rcsid[] = "$Id: ddpro.c,v 1.1 1993/08/30 23:26:47 manchek Exp $";
 static char pvmtxt[512];			/* scratch for error log */
 
 int dm_addhost();
 int dm_addhostack();
-int dm_bail();
 int dm_db();
 int dm_dback();
 int dm_delhost();
@@ -122,7 +133,6 @@ int (*netswitch[])() = {
 	dm_htcommit,
 	dm_slconf,
 	dm_startack,
-	dm_bail,
 	dm_task, dm_taskack,
 	dm_delhost, dm_delhostack,
 	dm_null,
@@ -144,7 +154,6 @@ char *dm_names[] = {
 	"dm_htcommit",
 	"dm_slconf",
 	"dm_startack",
-	"dm_bail",
 	"dm_task", "dm_taskack",
 	"dm_delhost", "dm_delhostack",
 	"dm_null",
@@ -422,6 +431,16 @@ hostid_new()
 }
 
 
+/*	dm_addhost()
+*
+*	(Master pvmd) gets request to add new hosts.
+*
+*	DM_ADDHOST(wid) {
+*		int count
+*		string name[count]
+*	}
+*/
+
 int
 dm_addhost(hp, mp)
 	struct hostd *hp;
@@ -435,8 +454,8 @@ dm_addhost(hp, mp)
 	int maxhostid = (tidhmask >> ffs(tidhmask) - 1);
 	char *name;
 
-/* XXX this should just queue the request, aargh */
-	if (newhosts) {
+/* XXX this should queue the request */
+	if (newhosts || pprime) {
 		struct mesg *mp2;
 
 		pvmlogerror("dm_addhost() already adding new hosts, oops\n");
@@ -474,6 +493,8 @@ dm_addhost(hp, mp)
 			newhosts = 0;
 			return 0;
 		}
+	/* XXX this is not right - shouldn't bind dtids to new hosts
+	   XXX until they're up and running. */
 		if (!(hh = hostid_new())) {
 			pvmlogerror("dm_addhost() out of hostids\n");
 			PVM_FREE(name);
@@ -522,6 +543,16 @@ dm_addhost(hp, mp)
 }
 
 
+/*	dm_addhostack()
+*
+*	Reply to DM_ADDHOST operation.
+*
+*	DM_ADDHOSTACK(wid_rtn) {
+*		int count			// or negative for error
+*		int status[count]	// status of each host
+*	}
+*/
+
 int
 dm_addhostack(hp, mp)
 	struct hostd *hp;
@@ -529,20 +560,8 @@ dm_addhostack(hp, mp)
 {
 	struct waitc *wp;
 
-	if (!(wp = wait_find(mp->m_wid))) {
-		if (mp->m_wid) {
-			sprintf(pvmtxt, "dm_addhostack() from t%x wid %d not found\n",
-					mp->m_src, mp->m_wid);
-			pvmlogerror(pvmtxt);
-		}
+	if (!(wp = wait_get(hp, mp, WT_ADDHOST)))
 		return 0;
-	}
-	if (wp->wa_on != hp->hd_hostpart) {
-		sprintf(pvmtxt, "dm_addhostack() from t%x wid %d wrong host?\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
-		return 0;
-	}
 	mp->m_ref++;
 	mp->m_dst = wp->wa_tid;
 	mp->m_cod = TM_ADDHOST;
@@ -554,10 +573,24 @@ dm_addhostack(hp, mp)
 
 /*	dm_spawn()
 *
-*	DM_SPAWN { ptid, file, flags, count, nargs, args }
 *	We try to fork/exec given number of tasks.  On first error, fill
 *	remainder of tid array with error code.  Return tid array to
 *	caller.
+*
+*	DM_SPAWN(wid) {
+*		int ptid
+*		string file
+*		int flags
+*		int count
+*		int nargs
+*		string argv[nargs]
+*		int outtid
+*		int outcod
+*		int trctid
+*		int trccod
+*		int nenv
+*		string env[nenv]
+*	}
 */
 
 int
@@ -571,6 +604,10 @@ dm_spawn(hp, mp)
 	int count;				/* number of tasks */
 	int nargs;				/* exec args */
 	char **av = 0;
+	int outtid, outcod;		/* stdout dst and code */
+	int trctid, trccod;		/* trace dst and code */
+	int nenv;				/* num of envars */
+	char **env = 0;			/* envars */
 	int *tids = 0;			/* list of tids from forkexec */
 	struct mesg *mp2;		/* reply message hdl */
 	struct task *tp = 0;
@@ -591,7 +628,7 @@ dm_spawn(hp, mp)
 	nargs += 2;
 	av = TALLOC(nargs + 1, char*, "argv");
 	av++;
-	bzero((char*)av, nargs * sizeof(char*));
+	BZERO((char*)av, nargs * sizeof(char*));
 	av[0] = file;
 	file = 0;
 	av[--nargs] = 0;
@@ -599,26 +636,69 @@ dm_spawn(hp, mp)
 		if (upkstralloc(mp, &av[i]))
 			goto bad;
 
+	if (upkuint(mp, &outtid)
+	|| upkuint(mp, &outcod)
+	|| upkuint(mp, &trctid)
+	|| upkuint(mp, &trccod)
+	|| upkuint(mp, &nenv))
+		goto bad;
+	BZERO((char*)(env = TALLOC((nenv + 1), char*, "env")),
+			(nenv + 1) * sizeof(char*));
+	for (i = 0; i < nenv; i++)
+		if (upkstralloc(mp, &env[i]))
+			goto bad;
+
 	if (count < 1)
 		goto bad;
 
 	tids = TALLOC(count, int, "tids");
 
+#if defined(IMA_PGON) || defined(IMA_I860) || defined(IMA_CM5)
+
+/* XXX only writing the first tid, oops. */
+	if (flags & PvmMppFront) {
+		/* spawn on service node */
+		if (err = forkexec(flags, av[0], av, nenv, env, &tp)) {
+			tids[0] = err;
+		} else {
+			tp->t_ptid = ptid;
+			tids[0] = tp->t_tid;
+		}
+	} else
+		mpp_load(flags, av[0], av, count, tids, ptid);
+
+#else /*defined(IMA_PGON) || defined(IMA_I860) || defined(IMA_CM5)*/
+
 	for (i = 0; i < count; i++) {
 		if (err) {
 			tids[i] = err;
 
 		} else {
-			if (err = forkexec(flags, av[0], av, &tp)) {
+			if (err = forkexec(flags, av[0], av, nenv, env, &tp)) {
 				tids[i] = err;
 
 			} else {
 				tp->t_ptid = ptid;
+				tp->t_outtid = outtid;
+				tp->t_outcod = outcod;
+				tp->t_trctid = trctid;
+				tp->t_trccod = trccod;
 				tids[i] = tp->t_tid;
+				if (outtid) {
+					mp2 = mesg_new(0);
+					mp2->m_cod = outcod;
+					mp2->m_dst = outtid;
+					pkint(mp2, tp->t_tid);
+					pkint(mp2, -2);
+					pkint(mp2, ptid);
+					sendmessage(mp2);
+				}
 			}
 		}
 	}
 
+#endif /*defined(IMA_PGON) || defined(IMA_I860) || defined(IMA_CM5)*/
+
 	mp2 = mesg_new(0);
 	pkint(mp2, count);
 	for (i = 0; i < count; i++)
@@ -645,10 +725,26 @@ cleanup:
 		av--;
 		PVM_FREE(av);
 	}
+	if (env) {
+		for (i = 0; i < nenv; i++)
+			if (env[i])
+				PVM_FREE(env[i]);
+		PVM_FREE(env);
+	}
 	return 0;
 }
 
 
+/*	dm_spawnack()
+*
+*	Reply to DM_SPAWN op.
+*
+*	DM_SPAWNACK(wid_rtn) {
+*		int count
+*		int tids[count]
+*	}
+*/
+
 int
 dm_spawnack(hp, mp)
 	struct hostd *hp;
@@ -661,18 +757,9 @@ dm_spawnack(hp, mp)
 	int v;
 	int err = 0;
 
-	if (!(wp = wait_find(mp->m_wid))) {
-		sprintf(pvmtxt, "dm_spawnack() from t%x wid %d not found\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
+	if (!(wp = wait_get(hp, mp, WT_SPAWN)))
 		return 0;
-	}
-	if (wp->wa_on != hp->hd_hostpart) {
-		sprintf(pvmtxt, "dm_spawnack() from t%x wid %d wrong host?\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
-		return 0;
-	}
+
 	wxp = (struct waitc_spawn*)wp->wa_spec;
 	upkint(mp, &rcnt);
 	v = wxp->w_veclen;
@@ -695,7 +782,7 @@ dm_spawnack(hp, mp)
 				break;
 		if (v < 0) {
 /* XXX aargh */
-			pvmlogerror("dm_spawnack() too many tids for result vector, we're hosed\n");
+			pvmlogerror("dm_spawnack() too many tids for result vector, hosing the request\n");
 			wait_delete(wp);
 			return 0;
 		}
@@ -717,6 +804,16 @@ dm_spawnack(hp, mp)
 }
 
 
+/*	dm_sendsig()
+*
+*	Request to send signal to local task.
+*
+*	DM_SENDSIG {
+*		int tid
+*		int signum
+*	}
+*/
+
 int
 dm_sendsig(hp, mp)
 	struct hostd *hp;
@@ -733,13 +830,19 @@ dm_sendsig(hp, mp)
 		return 0;
 	}
 	if (tp = task_find(tid)) {
+#if defined(IMA_I860) || defined(IMA_CM5)
+		(void)mpp_kill(tp, signum);
+
+#else
 		if (tp->t_pid)
 			(void)kill(tp->t_pid, signum);
+
 		else {
 			sprintf(pvmtxt, "dm_sendsig() signal for t%x scrapped (pid = 0)\n",
 					tid);
 			pvmlogerror(pvmtxt);
 		}
+#endif
 
 	} else {
 		sprintf(pvmtxt, "dm_sendsig() signal for t%x scrapped\n", tid);
@@ -749,6 +852,27 @@ dm_sendsig(hp, mp)
 }
 
 
+/*	dm_htupd()
+*
+*	Host table update phase 1 - (Non-master) pvmd is notified of new
+*	host table.
+*
+*	DM_HTUPD(wid) {
+*		int serial
+*		int master
+*		int console
+*		int count
+*		{
+*			int index
+*			string name
+*			string arch
+*			string hexipaddr
+*			int mtu
+*			int speed
+*		} [count]
+*	}
+*/
+
 int
 dm_htupd(hp, mp)
 	struct hostd *hp;
@@ -773,6 +897,9 @@ dm_htupd(hp, mp)
 
 	/* unpack new hosts and add to table */
 
+	/* XXX if we find a host already in table we should kill it with
+	   XXX hostfail and put the new one in its place */
+
 	upkint(mp, &count);
 	while (count-- > 0) {
 		upkint(mp, &hh);
@@ -782,6 +909,7 @@ dm_htupd(hp, mp)
 		upkstr(mp, buf, sizeof(buf));
 		hex_inadport(buf, &hp->hd_sad);
 		upkint(mp, &hp->hd_mtu);
+		upkint(mp, &hp->hd_speed);
 		ht_insert(newhosts, hp);
 		hd_unref(hp);
 	}
@@ -803,6 +931,13 @@ dm_htupd(hp, mp)
 }
 
 
+/*	dm_htupdack()
+*
+*	Reply to DM_HTUPD op.
+*
+*	DM_HTUPDACK(wid_rtn) { }
+*/
+
 int
 dm_htupdack(hp, mp)
 	struct hostd *hp;
@@ -810,18 +945,8 @@ dm_htupdack(hp, mp)
 {
 	struct waitc *wp;
 
-	if (!(wp = wait_find(mp->m_wid))) {
-		sprintf(pvmtxt, "dm_htupdack() from t%x wid %d not found\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
+	if (!(wp = wait_get(hp, mp, WT_HTUPD)))
 		return 0;
-	}
-	if (wp->wa_on != hp->hd_hostpart) {
-		sprintf(pvmtxt, "dm_htupdack() from t%x wid %d wrong host?\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
-		return 0;
-	}
 
 	/* is this is the last host checking in, send ht commit */
 
@@ -847,6 +972,77 @@ dm_htupdack(hp, mp)
 }
 
 
+/*	ht_diff()
+*
+*	Generates a host table containing entries in ht2 but not in ht1.
+*/
+
+struct htab *
+ht_diff(htp2, htp1)
+	struct htab *htp2;		/* more */
+	struct htab *htp1;		/* less */
+{
+	struct htab *htp;
+	int hh;
+	
+	htp = ht_new(1);
+	for (hh = htp2->ht_last; hh > 0; hh--)
+		if (htp2->ht_hosts[hh] && !htp2->ht_hosts[hh]->hd_err
+		&& (hh > htp1->ht_last || !htp1->ht_hosts[hh]))
+			ht_insert(htp, htp2->ht_hosts[hh]);
+	return htp;
+}
+
+
+/*	gotnewhosts()
+*
+*	Used to wake up any waitcs of kind WT_HOSTA.
+*	Sends a message containing count and list of new d-tids.
+*/
+
+gotnewhosts(htp2, htp1)
+	struct htab *htp2;		/* new host table */
+	struct htab *htp1;		/* old host table */
+{
+	struct mesg *mp;
+	struct htab *htp;
+	struct waitc *wp, *wp2;
+	int hh;
+
+	mp = 0;
+	for (wp = waitlist->wa_link; wp != waitlist; wp = wp2) {
+		wp2 = wp->wa_link;
+		if (wp->wa_kind == WT_HOSTA) {
+			if (!mp) {
+/* XXX shouldn't diff host tables and pack message several times, oops */
+				mp = mesg_new(0);
+				htp = ht_diff(htp2, htp1);
+				pkint(mp, htp->ht_cnt);
+				for (hh = htp->ht_last; hh > 0; hh--)
+					if (htp->ht_hosts[hh])
+						pkint(mp, htp->ht_hosts[hh]->hd_hostpart);
+				ht_free(htp);
+			}
+			mp->m_ref++;
+			mp->m_dst = wp->wa_tid;
+			mp->m_cod = wp->wa_dep;
+			sendmessage(mp);
+			wait_delete(wp);
+		}
+	}
+	if (mp)
+		mesg_unref(mp);
+	return 0;
+}
+
+
+/*	dm_htcommit()
+*
+*	Host table update phase 2 - commit to new host table.
+*
+*	DM_HTCOMMIT { }
+*/
+
 int
 dm_htcommit(hp, mp)
 	struct hostd *hp;
@@ -870,6 +1066,8 @@ dm_htcommit(hp, mp)
 					htp->ht_serial, hosts->ht_serial);
 			pvmlogerror(pvmtxt);
 		}
+
+		gotnewhosts(hosts, htp);
 		ht_free(htp);
 		runstate = PVMDNORMAL;
 
@@ -880,14 +1078,60 @@ dm_htcommit(hp, mp)
 }
 
 
+/*	dm_slconf()
+*
+*	Pvmd gets config info (from master).
+*	One of these should arrive before a pvmd is fully up, even
+*	before the host table update, to set private params.
+*	More may be used later to set random knobs.
+*
+*	DM_SLCONF {
+*		{
+*			int fieldtype		// from ddpro.h
+*			string value
+*		} []					// implied
+*	}
+*/
+
 int
 dm_slconf(hp, mp)
 	struct hostd *hp;
 	struct mesg *mp;
 {
+	int t;				/* field type */
+	char *s;
+
 	mp = mp;
 	hp = hp;
-	pvmlogerror("dm_slconf() write me XXX\n");
+
+	while (!upkint(mp, &t) && !upkstralloc(mp, &s)) {
+		switch (t) {
+
+		case DM_SLCONF_EP:
+			epaths = pathsep(s, ':', myarchname, '%');
+			if (debugmask & PDMSTARTUP) {
+				sprintf(pvmtxt, "dm_slconf() ep<%s>\n", s);
+				pvmlogerror(pvmtxt);
+			}
+			break;
+
+		case DM_SLCONF_BP:
+			debugger = s;
+			if (debugmask & PDMSTARTUP) {
+				sprintf(pvmtxt, "dm_slconf() bp<%s>\n", s);
+				pvmlogerror(pvmtxt);
+			}
+			break;
+
+		default:
+			if (strlen(s) > sizeof(pvmtxt)-50)
+				s[sizeof(pvmtxt)-50] = 0;
+			sprintf(pvmtxt, "dm_slconf() ? type %d val <%s>\n", t, s);
+			pvmlogerror(pvmtxt);
+			PVM_FREE(s);
+			break;
+		}
+	}
 	return 0;
 }
 
@@ -896,6 +1140,24 @@ dm_slconf(hp, mp)
 *
 *	This is how the new host table info gets from pvmd' back to
 *	pvmd[master].
+*
+*	DM_STARTACK(wid_rtn) {
+*		int count
+*		{
+*			int index
+*			int errcode			// zero if ok
+*			{					// if errcode == 0
+*				string arch
+*				string hexipaddr
+*				int mtu
+*				{					// to be fwd in DM_SLCONF message
+*					int fieldtype	// from ddpro.h
+*					string value
+*				} []				// implied
+*				int 0				// marks end of options
+*			}
+*		} [count]
+*	}
 */
 
 int
@@ -905,21 +1167,33 @@ dm_startack(hp, mp)
 {
 	struct waitc *wp;		/* wait context on pvmd' */
 	int count;				/* num of new hosts */
+	int happy;				/* num of happy new hosts */
 	int hh;
 	char buf[16];			/* for converting sockaddr */
 	struct waitc *wp2;		/* 'seed' waitc for peer wait on htupdack */
 	struct waitc *wp3;
+	struct mesg *mp2;
+	int t;
+	char *s;
 
-	wp = wait_find(mp->m_wid);
+	finack_to_host(hp);
+
+	if (!(wp = wait_get(hp, mp, WT_HOSTSTART)))
+		return 0;
 
 	if (!newhosts) {
 		pvmlogerror("dm_startack() no new host table?\n");
 		return 0;
 	}
 
-	/* unpack host table info (from pvmd'), drop into newhosts */
+	/*
+	* unpack host table info (from pvmd'), put stats or err codes in newhosts.
+	* forward options to new slaves with DM_SLCONF message.
+	* if one or more started ok, do host-table update.
+	*/
 
 	upkint(mp, &count);
+	happy = 0;
 	while (count-- > 0) {
 		upkint(mp, &hh);
 		hp = newhosts->ht_hosts[hh];
@@ -929,6 +1203,20 @@ dm_startack(hp, mp)
 			upkstr(mp, buf, sizeof(buf));
 			hex_inadport(buf, &hp->hd_sad);
 			upkint(mp, &hp->hd_mtu);
+
+			if (!happy++)
+				runstate = PVMDHTUPD;
+
+			mp2 = mesg_new(0);
+			mp2->m_cod = DM_SLCONF;
+			while (!upkint(mp, &t) && t) {
+				pkint(mp2, t);
+				upkstralloc(mp, &s);
+				pkstr(mp2, s);
+				PVM_FREE(s);
+			}
+			mp2->m_dst = hp->hd_hostpart | TIDPVMD;
+			sendmessage(mp2);
 		}
 	}
 
@@ -939,49 +1227,40 @@ dm_startack(hp, mp)
 
 	wp2 = wait_new(WT_HTUPD);
 
-	/* compose reply to waiter */
+	/*
+	* compose reply to waiter.  attach it to a wait context to go
+	* when the host table update is completed.
+	*/
 
-	if (wp) {
-		mp = mesg_new(0);
-		mp->m_dst = wp->wa_tid;
-		mp->m_cod = DM_ADDHOSTACK;
-		mp->m_wid = wp->wa_dep;
-		pkint(mp, newhosts->ht_cnt);
+	mp = mesg_new(0);
+	mp->m_dst = wp->wa_tid;
+	mp->m_cod = DM_ADDHOSTACK;
+	mp->m_wid = wp->wa_dep;
+	pkint(mp, newhosts->ht_cnt);
 /* XXX danger, the dtids may be out of order w/r/t the names sent
    XXX if the hostid wraps */
-		for (hh = 1; hh <= newhosts->ht_last; hh++)
-			if (hp = newhosts->ht_hosts[hh])
-				pkint(mp, (hp->hd_err ? hp->hd_err : hp->hd_hostpart));
-		wp2->wa_mesg = mp;
-		wait_delete(wp);
-		wp = 0;
+	for (hh = 1; hh <= newhosts->ht_last; hh++)
+		if (hp = newhosts->ht_hosts[hh])
+			pkint(mp, (hp->hd_err ? hp->hd_err : hp->hd_hostpart));
+	wp2->wa_mesg = mp;
+	wait_delete(wp);
+	wp = 0;
 
-	} else {
-		sprintf(pvmtxt, "dm_startack() from t%x wid %d not found\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
-pvmlogerror("AARGH! AARGH! XXX\n");
-return 0;
-	}
+	/* if at least one working new host */
 
-	/* check at least one working new host */
-
-	count = 0;
-	for (hh = newhosts->ht_last; hh > 0; hh--)
-		if ((hp = newhosts->ht_hosts[hh]) && !hp->hd_err)
-			count++;
-
-	if (count) {
-		runstate = PVMDHTUPD;
-
-	/* send each new host the host table data for current + new hosts */
+	if (happy) {
+		
+		/*
+		* create host table update message containing all current hosts
+		* plus happy new ones.
+		*/
 
 		mp = mesg_new(0);
 		mp->m_cod = DM_HTUPD;
 		pkint(mp, newhosts->ht_serial);
 		pkint(mp, hosts->ht_master);
 		pkint(mp, hosts->ht_cons);
-		pkint(mp, hosts->ht_cnt + count);
+		pkint(mp, hosts->ht_cnt + happy);
 		for (hh = hosts->ht_last; hh > 0; hh--)
 			if ((hp = hosts->ht_hosts[hh])) {
 				pkint(mp, hh);
@@ -989,6 +1268,7 @@ return 0;
 				pkstr(mp, hp->hd_arch);
 				pkstr(mp, inadport_hex(&hp->hd_sad));
 				pkint(mp, hp->hd_mtu);
+				pkint(mp, hp->hd_speed);
 			}
 		for (hh = newhosts->ht_last; hh > 0; hh--)
 			if ((hp = newhosts->ht_hosts[hh]) && !hp->hd_err) {
@@ -997,9 +1277,12 @@ return 0;
 				pkstr(mp, hp->hd_arch);
 				pkstr(mp, inadport_hex(&hp->hd_sad));
 				pkint(mp, hp->hd_mtu);
+				pkint(mp, hp->hd_speed);
 			}
 
-	/* send to each new host */
+		/*
+		* send to each new host
+		*/
 
 		for (hh = newhosts->ht_last; hh > 0; hh--)
 			if ((hp = newhosts->ht_hosts[hh]) && !hp->hd_err) {
@@ -1015,14 +1298,16 @@ return 0;
 			}
 		mesg_unref(mp);
 
-	/* send each old host the host table data for new hosts */
+		/*
+		* create host table update message containing happy new ones.
+		*/
 
 		mp = mesg_new(0);
 		mp->m_cod = DM_HTUPD;
 		pkint(mp, newhosts->ht_serial);
 		pkint(mp, newhosts->ht_master);
 		pkint(mp, newhosts->ht_cons);
-		pkint(mp, count);
+		pkint(mp, happy);
 		for (hh = newhosts->ht_last; hh > 0; hh--)
 			if ((hp = newhosts->ht_hosts[hh]) && !hp->hd_err) {
 				pkint(mp, hh);
@@ -1030,9 +1315,12 @@ return 0;
 				pkstr(mp, hp->hd_arch);
 				pkstr(mp, inadport_hex(&hp->hd_sad));
 				pkint(mp, hp->hd_mtu);
+				pkint(mp, hp->hd_speed);
 			}
 
-	/* send to each old host */
+		/*
+		* send to each old host
+		*/
 
 		for (hh = hosts->ht_last; hh > 0; hh--)
 			if (hh != hosts->ht_local && (hp = hosts->ht_hosts[hh])) {
@@ -1048,12 +1336,15 @@ return 0;
 			}
 		mesg_unref(mp);
 
-	/* update our host table */
+		/*
+		* update our host table
+		*/
 
 	/* XXX returning to normal state right here is a hack, we should
 	   XXX wait for all the ht upd acks to come back but we need the
 	   XXX regular message service, hostfail entry, etc. */
 
+		gotnewhosts(newhosts, hosts);
 		for (hh = newhosts->ht_last; hh > 0; hh--)
 			if ((hp = newhosts->ht_hosts[hh]) && !hp->hd_err)
 				ht_insert(hosts, hp);
@@ -1078,24 +1369,14 @@ return 0;
 }
 
 
-int
-dm_bail(hp, mp)
-	struct hostd *hp;
-	struct mesg *mp;
-{
-	if (hp == hosts->ht_hosts[hosts->ht_master]) {
-		sprintf(pvmtxt, "dm_bail() warning from master (%s), halting...\n",
-				hp->hd_name);
-		pvmlogerror(pvmtxt);
-		runstate = PVMDHALTING;
-	} else {
-		sprintf(pvmtxt, "dm_bail() from t%x (not master)?\n",
-				mp->m_src);
-		pvmlogerror(pvmtxt);
-	}
-	return 0;
-}
-
+/*	dm_task()
+*
+*	(Remote or local) pvmd requests information about local task(s).
+*
+*	DM_TASK(wid) {
+*		int where			// 0 for all, hostpart, or full tid
+*	}
+*/
 
 int
 dm_task(hp, mp)
@@ -1104,27 +1385,59 @@ dm_task(hp, mp)
 {
 	struct task *tp;
 	struct mesg *mp2;
+	int where;
 
 	hp = hp;
 
+	if (upkuint(mp, &where)) {
+		pvmlogerror("dm_task() bad msg format\n");
+		return 0;
+	}
+
 	/* pack list of local tasks and reply to waiter */
 
 	mp2 = mesg_new(0);
 	mp2->m_dst = mp->m_src;
 	mp2->m_cod = DM_TASKACK;
 	mp2->m_wid = mp->m_wid;
-	for (tp = locltasks->t_link; tp != locltasks; tp = tp->t_link) {
-		pkint(mp2, tp->t_tid);
-		pkint(mp2, tp->t_ptid);
-		pkint(mp2, myhostpart);
-		pkint(mp2, tp->t_flag);
-		pkstr(mp2, (tp->t_a_out ? tp->t_a_out : ""));
+	if (where & tidlmask) {
+		if (tp = task_find(where)) {
+			pkint(mp2, tp->t_tid);
+			pkint(mp2, tp->t_ptid);
+			pkint(mp2, myhostpart);
+			pkint(mp2, tp->t_flag);
+			pkstr(mp2, tp->t_a_out ? tp->t_a_out : "");
+		}
+
+	} else {
+		for (tp = locltasks->t_link; tp != locltasks; tp = tp->t_link) {
+			pkint(mp2, tp->t_tid);
+			pkint(mp2, tp->t_ptid);
+			pkint(mp2, myhostpart);
+			pkint(mp2, tp->t_flag);
+			pkstr(mp2, (tp->t_a_out ? tp->t_a_out : ""));
+		}
 	}
 	sendmessage(mp2);
 	return 0;
 }
 
 
+/*	dm_taskack()
+*
+*	Reply to DM_TASK op.
+*
+*	DM_TASKACK(wid_rtn) {
+*		{
+*			int tid
+*			int ptid
+*			int hostpart
+*			int flag
+*			string a_out
+*		} []				// implied
+*	}
+*/
+
 int
 dm_taskack(hp, mp)
 	struct hostd *hp;
@@ -1135,31 +1448,21 @@ dm_taskack(hp, mp)
 	int i;
 	char *p;
 
-	if (!(wp = wait_find(mp->m_wid))) {
-		sprintf(pvmtxt, "dm_taskack() from t%x wid %d not found\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
+	if (!(wp = wait_get(hp, mp, WT_TASK)))
 		return 0;
-	}
-	if (wp->wa_on != hp->hd_hostpart) {
-		sprintf(pvmtxt, "dm_taskack() from t%x wid %d wrong host?\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
-		return 0;
-	}
 
 	/* append data to waiting message */
 
 	mp2 = wp->wa_mesg;
 	while (!upkint(mp, &i)) {
+		pkint(mp2, i);			/* tid */
+		upkint(mp, &i);			/* ptid */
 		pkint(mp2, i);
-		upkint(mp, &i);
+		upkint(mp, &i);			/* host */
 		pkint(mp2, i);
-		upkint(mp, &i);
+		upkint(mp, &i);			/* flag */
 		pkint(mp2, i);
-		upkint(mp, &i);
-		pkint(mp2, i);
-		upkstralloc(mp, &p);
+		upkstralloc(mp, &p);	/* a.out name */
 		pkstr(mp2, p);
 		PVM_FREE(p);
 	}
@@ -1176,6 +1479,16 @@ dm_taskack(hp, mp)
 }
 
 
+/*	dm_delhost()
+*
+*	(Master pvmd) gets request to delete hosts.
+*
+*	DM_DELHOST(wid) {
+*		int count
+*		string names[count]
+*	}
+*/
+
 int
 dm_delhost(hp, mp)
 	struct hostd *hp;
@@ -1236,7 +1549,7 @@ dm_delhost(hp, mp)
 					ht_insert(ht_del, hp);
 					ht_delete(ht_save, hp);
 					pkint(mp3, hp->hd_hostpart);
-					hp->hd_flag |= HF_SHUTDOWN;
+					fin_to_host(hp);
 					pkint(mp2, 0);
 				}
 			}
@@ -1245,20 +1558,6 @@ dm_delhost(hp, mp)
 			pkint(mp2, PvmNoHost);
 	}
 
-	/*
-	* send BAIL message to all hosts in ht_del set
-	*/
-
-	mp = mesg_new(0);
-	mp->m_cod = DM_BAIL;
-	for (hh = ht_del->ht_last; hh > 0; hh--)
-		if (hp = ht_del->ht_hosts[hh]) {
-			mp->m_dst = hp->hd_hostpart | TIDPVMD;
-			mp->m_ref++;
-			sendmessage(mp);
-		}
-	mesg_unref(mp);
-
 	/*
 	* send HTDEL message to all hosts in ht_save set except us
 	*/
@@ -1281,6 +1580,16 @@ dm_delhost(hp, mp)
 }
 
 
+/*	dm_delhostack()
+*
+*	Reply to DM_DELHOST operation.
+*
+*	DM_DELHOSTACK(wid_rtn) {
+*		int count			// or negative for error
+*		int status[count]	// status of each host
+*	}
+*/
+
 int
 dm_delhostack(hp, mp)
 	struct hostd *hp;
@@ -1288,18 +1597,9 @@ dm_delhostack(hp, mp)
 {
 	struct waitc *wp;
 
-	if (!(wp = wait_find(mp->m_wid))) {
-		sprintf(pvmtxt, "dm_delhostack() from t%x wid %d not found\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
+	if (!(wp = wait_get(hp, mp, WT_DELHOST)))
 		return 0;
-	}
-	if (wp->wa_on != hp->hd_hostpart) {
-		sprintf(pvmtxt, "dm_delhostack() from t%x wid %d wrong host?\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
-		return 0;
-	}
+
 	mp->m_ref++;
 	mp->m_dst = wp->wa_tid;
 	mp->m_cod = TM_DELHOST;
@@ -1309,6 +1609,13 @@ dm_delhostack(hp, mp)
 }
 
 
+/*	dm_null()
+*
+*	No-op message.
+*
+*	DM_NULL { }
+*/
+
 int
 dm_null(hp, mp)
 	struct hostd *hp;
@@ -1324,22 +1631,69 @@ dm_null(hp, mp)
 }
 
 
+/*	dm_taskout()
+*
+*	Stdout data from a task.
+*
+*	DM_TASKOUT {
+*		int tid
+*		int length
+*		char data[length]
+*	}
+*/
+
 int
 dm_taskout(hp, mp)
 	struct hostd *hp;
 	struct mesg *mp;
 {
 	int tid;
+	int l;
+	char *p, *q, c;
+	char buf2[32];
+	char buf[4100];		 /* a bit bigger than in pvmd.c`loclstout() */
 
 	hp = hp;
-	upkuint(mp, &tid);
-	sprintf(pvmtxt, "[t%x] ", tid);
-	upkstr(mp, pvmtxt + strlen(pvmtxt), sizeof(pvmtxt) - strlen(pvmtxt));
-	pvmlogerror(pvmtxt);
+
+	if (upkuint(mp, &tid) || upkint(mp, &l) || l < 1)
+		return 0;
+
+	/* unpack data, leaving room at head of buffer */
+
+	p = buf + 32;
+	if (l > sizeof(buf) - (p - buf) - 2)
+		l = sizeof(buf) - (p - buf) - 2;
+	byteupk(mp, p, l, 1, 1);
+
+	/* ends with "\n\0" */
+	if (p[l - 1] != '\n')
+		p[l++] = '\n';
+	p[l] = 0;
+
+	sprintf(buf2, "[t%x] ", tid);
+	l = strlen(buf2);
+	while (*p) {
+		for (q = p; *q++ != '\n'; ) ;
+		c = *q;
+		*q = 0;
+		BCOPY(buf2, p - l, l);
+		pvmlogerror(p - l);
+		*q = c;
+		p = q;
+	}
 	return 0;
 }
 
 
+/*	dm_pstat()
+*
+*	(Remote or local) pvmd requests status of a local task.
+*
+*	DM_PSTAT(wid) {
+*		int tid
+*	}
+*/
+
 int
 dm_pstat(hp, mp)
 	struct hostd *hp;
@@ -1364,6 +1718,15 @@ dm_pstat(hp, mp)
 }
 
 
+/*	dm_pstatack()
+*
+*	Reply to DM_PSTAT op.
+*
+*	DM_PSTATACK(wid_rtn) {
+*		int status
+*	}
+*/
+
 int
 dm_pstatack(hp, mp)
 	struct hostd *hp;
@@ -1371,18 +1734,9 @@ dm_pstatack(hp, mp)
 {
 	struct waitc *wp;
 
-	if (!(wp = wait_find(mp->m_wid))) {
-		sprintf(pvmtxt, "dm_pstatack() from t%x wid %d not found\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
+	if (!(wp = wait_get(hp, mp, 0)))
 		return 0;
-	}
-	if (wp->wa_on != hp->hd_hostpart) {
-		sprintf(pvmtxt, "dm_pstatack() from t%x wid %d wrong host?\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
-		return 0;
-	}
+
 	if (wp->wa_kind == WT_PSTAT)
 		mp->m_cod = TM_PSTAT;
 	else
@@ -1395,6 +1749,13 @@ dm_pstatack(hp, mp)
 }
 
 
+/*	dm_halt()
+*
+*	Task has requested whole machine to halt.
+*
+*	DM_HALT { }
+*/
+
 int
 dm_halt(hp, mp)
 	struct hostd *hp;
@@ -1405,20 +1766,28 @@ dm_halt(hp, mp)
 	sprintf(pvmtxt, "dm_halt() from (%s), halting...\n",
 			hp->hd_name);
 	pvmlogerror(pvmtxt);
-	mp = mesg_new(0);
-	mp->m_cod = DM_BAIL;
 	for (hh = hosts->ht_last; hh > 0; hh--) {
 		if (hh == hosts->ht_local || !(hp = hosts->ht_hosts[hh]))
 			continue;
-		mp->m_ref++;
-		mp->m_dst = hp->hd_hostpart | TIDPVMD;
-		sendmessage(mp);
+		finack_to_host(hp);
 	}
 	runstate = PVMDHALTING;
 	return 0;
 }
 
 
+/*	dm_mca()
+*
+*	Remote pvmd defines a new (single-use) multicast address, to be used
+*	by a subsequent message.
+*
+*	DM_MCA {
+*		int gtid			// multicast tid
+*		int count			// number of addresses
+*		int tids[count]		// addresses
+*	}
+*/
+
 int
 dm_mca(hp, mp)
 	struct hostd *hp;
@@ -1448,6 +1817,16 @@ dm_mca(hp, mp)
 }
 
 
+/*	dm_notify()
+*
+*	Remote pvmd requests to be notified on an event.
+*
+*	DM_NOTIFY(wid) {
+*		int what			// event type XXX currenty PvmTaskExit
+*		int tid				// address
+*	}
+*/
+
 int
 dm_notify(hp, mp)
 	struct hostd *hp;
@@ -1479,6 +1858,13 @@ dm_notify(hp, mp)
 }
 
 
+/*	dm_notifyack()
+*
+*	Reply to DM_NOTIFY op.
+*
+*	DM_NOTIFYACK(wid_rtn) { }
+*/
+
 int
 dm_notifyack(hp, mp)
 	struct hostd *hp;
@@ -1488,12 +1874,8 @@ dm_notifyack(hp, mp)
 
 	hp = hp;
 
-	if (!(wp = wait_find(mp->m_wid))) {
-		sprintf(pvmtxt, "dm_notifyack() from t%x wid %d not found\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
+	if (!(wp = wait_get((struct hostd*)0, mp, 0)))
 		return 0;
-	}
 
 	mp = mesg_new(0);
 	mp->m_dst = wp->wa_tid;
@@ -1506,6 +1888,18 @@ dm_notifyack(hp, mp)
 }
 
 
+/*	dm_db()
+*
+*	(Remote or local) pvmd requests a database op.
+*
+*	DM_DB(wid) {
+*		int opcode		// insert, delete, lookup
+*		string name
+*		int index
+*		int data
+*	}
+*/
+
 int
 dm_db(hp, mp)
 	struct hostd *hp;
@@ -1565,6 +1959,16 @@ badformat:
 }
 
 
+/*	dm_dback()
+*
+*	Reply to DM_DB op.
+*
+*	DM_DBACK(wid_rtn) {
+*		int index		// and status
+*		int data		// if 'lookup'
+*	}
+*/
+
 int
 dm_dback(hp, mp)
 	struct hostd *hp;
@@ -1572,18 +1976,9 @@ dm_dback(hp, mp)
 {
 	struct waitc *wp;
 
-	if (!(wp = wait_find(mp->m_wid))) {
-		sprintf(pvmtxt, "dm_dback() from t%x wid %d not found\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
+	if (!(wp = wait_get(hp, mp, WT_DB)))
 		return 0;
-	}
-	if (wp->wa_on != hp->hd_hostpart) {
-		sprintf(pvmtxt, "dm_dback() from t%x wid %d wrong host?\n",
-				mp->m_src, mp->m_wid);
-		pvmlogerror(pvmtxt);
-		return 0;
-	}
+
 	mp->m_ref++;
 	mp->m_dst = wp->wa_tid;
 	mp->m_cod = TM_DB;
