@@ -1,11 +1,11 @@
 
 /*
- *           PVM 3.0:  Parallel Virtual Machine System 3.0
+ *           PVM 3.2:  Parallel Virtual Machine System 3.2
  *               University of Tennessee, Knoxville TN.
  *           Oak Ridge National Laboratory, Oak Ridge TN.
  *                   Emory University, Atlanta GA.
  *      Authors:  A. L. Beguelin, J. J. Dongarra, G. A. Geist,
- *          R. J. Manchek, B. K. Moore, and V. S. Sunderam
+ *    W. C. Jiang, R. J. Manchek, B. K. Moore, and V. S. Sunderam
  *                   (C) 1992 All Rights Reserved
  *
  *                              NOTICE
@@ -22,7 +22,7 @@
  * purpose.  This software is provided ``as is'' without express or
  * implied warranty.
  *
- * PVM 3.0 was funded in part by the U.S. Department of Energy, the
+ * PVM 3.2 was funded in part by the U.S. Department of Energy, the
  * National Science Foundation and the State of Tennessee.
  */
 
@@ -31,15 +31,20 @@
  *
  *	Libpvm core for unix environment.
  *
-$Log$
+$Log: lpvm.c,v $
+ * Revision 1.1  1993/08/30  23:26:48  manchek
+ * Initial revision
+ *
  */
 
-#define	READWHOLEMESSAGE
 
 #include <stdio.h>
 #ifdef IMA_BSD386
 #include <machine/endian.h>
 #endif
+#ifdef IMA_LINUX
+#include <endian.h>
+#endif
 #include <rpc/types.h>
 #include <rpc/xdr.h>
 #include <sys/time.h>
@@ -58,7 +63,7 @@ $Log$
 #endif
 #include <errno.h>
 #include <signal.h>
-#include "pvm3.h"
+#include <pvm3.h>
 #include "global.h"
 #include "tdpro.h"
 #include "ddpro.h"
@@ -66,10 +71,8 @@ $Log$
 #include "pvmfrag.h"
 #include "pvmumbuf.h"
 #include "listmac.h"
-
-#ifndef	PVMDFILE
-#define	PVMDFILE	"pvm3/lib/pvmd"
-#endif
+#include "tvdefs.h"
+#include "bfunc.h"
 
 /* task debug mask */
 
@@ -78,9 +81,6 @@ $Log$
 #define	TDMSELECT	4		/* select and fd sets */
 #define	TDMROUTE	8		/* message route control */
 
-
-#define	TVCLEAR(tvp)	((tvp)->tv_sec = (tvp)->tv_usec = 0)
-
 /*
 *	task-task or task-pvmd control block
 */
@@ -114,7 +114,7 @@ char *getenv();
 extern struct encvec *enctovec();
 extern char *inadport_decimal();
 extern char *inadport_hex();
-
+char *pvmgetpvmd();
 
 
 /***************
@@ -166,6 +166,12 @@ char *pvm_errlist[] = {					/* error messages for -pvm_errno */
 	"Duplicate entry",
 };
 
+int pvmautoerr = 1;						/* whether to auto print err msgs */
+int pvmcouttid = 0;						/* child stdout dst and code */
+int pvmcoutcod = 0;
+int pvmctrctid = 0;						/* child trace dst and code */
+int pvmctrccod = 0;
+int pvmfrgsiz = UDPMAXLEN;				/* message frag length (to pack) */
 int pvmmyptid = -1;						/* parent task id */
 int pvmmytid = -1;						/* this task id */
 int pvmmyupid = -1;						/* process unix pid */
@@ -176,6 +182,10 @@ int pvm_nerr = sizeof(pvm_errlist)
 			/sizeof(pvm_errlist[0]);	/* exported num of errors */
 struct umbuf *pvmrxlist = 0;			/* not-recvd msg list */
 int pvmtidhmask = TIDHOST;				/* mask for host field of tids */
+int pvmouttid = 0;						/* stdout dst and code */
+int pvmoutcod = 0;
+int pvmtrctid = 0;						/* trace dst and code */
+int pvmtrccod = 0;
 
 
 /***************
@@ -183,7 +193,7 @@ int pvmtidhmask = TIDHOST;				/* mask for host field of tids */
  **           **
  ***************/
 
-static char rcsid[] = "$Id$";
+static char rcsid[] = "$Id: lpvm.c,v 1.1 1993/08/30 23:26:48 manchek Exp $";
 static int debugmask = 0;				/* which debugging info */
 static int mxfersingle = 1;				/* mxfer returns after single frag */
 static char pvmtxt[512];				/* scratch for error log */
@@ -351,10 +361,10 @@ ttpcb_new()
 	struct ttpcb *pcbp;
 
 	if (pcbp = TALLOC(1, struct ttpcb, "tpcb")) {
-		bzero((char*)pcbp, sizeof(struct ttpcb));
+		BZERO((char*)pcbp, sizeof(struct ttpcb));
 		pcbp->tt_fd = -1;
 		pcbp->tt_rxfrag = TALLOC(1, struct umbuf, "umb");
-		bzero((char*)pcbp->tt_rxfrag, sizeof(struct umbuf));
+		BZERO((char*)pcbp->tt_rxfrag, sizeof(struct umbuf));
 		pcbp->tt_rxfrag->ub_link = pcbp->tt_rxfrag->ub_rlink = pcbp->tt_rxfrag;
 	}
 	return pcbp;
@@ -638,6 +648,7 @@ mctl(up)
 	struct umbuf *up2;
 	int i;
 	int tid;
+	static int linger[2] = { 1, 60 };	/* XXX arbitrary time */
 
 	switch (up->ub_cod) {
 
@@ -662,13 +673,17 @@ mctl(up)
 
 				} else {
 					pcbp->tt_state = TTOPEN;
-#if 0
 #ifndef NOSOCKOPT
+#if 0
 					i = 1;
 					if (setsockopt(pcbp->tt_fd, IPPROTO_TCP, TCP_NODELAY,
 							(char*)&i, sizeof(int)) == -1)
 						pvmlogperror("mctl() setsockopt");
-#endif
+#endif /*0*/
+					if (setsockopt(pcbp->tt_fd, SOL_SOCKET, SO_LINGER,
+							(char*)linger, sizeof(linger)) == -1)
+						pvmlogperror("mctl() setsockopt");
+#endif /*NOSOCKOPT*/
 					if ((i = fcntl(pcbp->tt_fd, F_GETFL, 0)) == -1)
 						pvmlogperror("mctl() fcntl");
 					else {
@@ -679,12 +694,11 @@ mctl(up)
 #endif
 						(void)fcntl(pcbp->tt_fd, F_SETFL, i);
 					}
-#endif/*0*/
 					pvm_fd_add(pcbp->tt_fd, 1);
 				}
 
 			} else {
-				sprintf(pvmtxt, "mctl() CONREQ from t% but state=%d ?\n",
+				sprintf(pvmtxt, "mctl() CONREQ from t%x but state=%d ?\n",
 						up->ub_src, pcbp->tt_state);
 				pvmlogerror(pvmtxt);
 			}
@@ -773,14 +787,18 @@ mctl(up)
 
 						} else {
 							pcbp->tt_state = TTOPEN;
-#if 0
 #ifndef NOSOCKOPT
+#if 0
 							i = 1;
 							if (setsockopt(pcbp->tt_fd, IPPROTO_TCP,
 									TCP_NODELAY, (char*)&i, sizeof(int))
 							== -1)
 								pvmlogperror("mctl() setsockopt");
-#endif
+#endif /*0*/
+							if (setsockopt(pcbp->tt_fd, SOL_SOCKET, SO_LINGER,
+									(char*)linger, sizeof(linger)) == -1)
+								pvmlogperror("mctl() setsockopt");
+#endif /*NOSOCKOPT*/
 							if ((i = fcntl(pcbp->tt_fd, F_GETFL, 0)) == -1)
 								pvmlogperror("mctl() fcntl");
 							else {
@@ -791,7 +809,6 @@ mctl(up)
 #endif
 								(void)fcntl(pcbp->tt_fd, F_SETFL, i);
 							}
-#endif/*0*/
 							pvm_fd_add(pcbp->tt_fd, 1);
 						}
 
@@ -848,41 +865,11 @@ mctl(up)
 }
 
 
-#ifdef	READWHOLEMESSAGE
-/*	tread()
-*
-*	Tenacious read.  Just like read(2) except that it will keep trying
-*	to get those n bytes until there really aren't any more.
-*	Returns the number it got or -1 if error.
-*/
-
-static int
-tread(d, b, n)
-	int d;
-	char *b;
-	int n;
-{
-	int e;
-	int x = n;
-
-	while (x > 0) {
-		if ((e = read(d, b, x)) == -1)
-			return e;
-		if (!e)
-			return n - x;
-		b += e;
-		x -= e;
-	}
-	return n - x;
-}
-#endif	/*READWHOLEMESSAGE*/
-
-
 /*	mxinput()
 *
 *	Input from a connection.
 *	Reassemble packets and messages and add them to pvmrxlist.
-*	Returns the number of messages added to pvmrxlist,
+*	Returns >= 0 the number of messages added to pvmrxlist,
 *	or negative on error.
 */
 
@@ -890,50 +877,77 @@ int
 mxinput(pcbp)
 	struct ttpcb *pcbp;
 
-#ifdef	READWHOLEMESSAGE
 {
 	int gotem = 0;			/* num msgs added to pvmrxlist */
-	struct frag *fp;
+	struct frag *fp;		/* shadows pcbp->tt_rxf */
+	struct frag *fp2;
+	int n;					/* bytes received */
+	int m;					/* length of fragment */
+	struct umbuf *rxup;		/* message containing this frag */
 	char *cp;				/* gp */
 	int src;
 	int ff;
-	int n;
-	struct umbuf *rxup;
-	char header[TDFRAGHDR];
-	int len;
 
-	while (!gotem) {
+	if (!pcbp->tt_rxf)
+		pcbp->tt_rxf = fr_new(pvmfrgsiz);
+	fp = pcbp->tt_rxf;
 
-		n = tread(pcbp->tt_fd, header, TDFRAGHDR);
-		if (n != TDFRAGHDR) {
-			if (n == -1)
-				pvmlogperror("mxinput() tread");
-			else {
-				if (debugmask & TDMPACKET) {
-					sprintf(pvmtxt, "mxinput() t%x read EOF\n", pcbp->tt_tid);
-					pvmlogerror(pvmtxt);
-				}
-			}
-			return PvmSysErr;
+	/* read fragment header separately from body */
+
+	n = (fp->fr_len < TDFRAGHDR) ? 0 : pvmget32(fp->fr_dat + 8);
+	n += TDFRAGHDR - fp->fr_len;
+	if (debugmask & TDMPACKET) {
+		sprintf(pvmtxt, "mxinput() pcb t%x fr_len=%d fr_dat=+%d n=%d\n",
+				pcbp->tt_tid, fp->fr_len, fp->fr_dat - fp->fr_buf, n);
+		pvmlogerror(pvmtxt);
+	}
+	n = read(pcbp->tt_fd, fp->fr_dat + fp->fr_len, n);
+	if (debugmask & TDMPACKET) {
+		sprintf(pvmtxt, "mxinput() read=%d\n", n);
+		pvmlogerror(pvmtxt);
+	}
+
+	/* deal with errors and closes */
+
+	if (n == -1 && errno != EWOULDBLOCK && errno != EINTR) {
+		if (debugmask & TDMPACKET) {
+			sprintf(pvmtxt, "mxinput() t%x read", pcbp->tt_tid);
+			pvmlogperror(pvmtxt);
 		}
+		return PvmSysErr;
+	}
+	if (!n) {
+		if (debugmask & TDMPACKET) {
+			sprintf(pvmtxt, "mxinput() t%x read EOF\n", pcbp->tt_tid);
+			pvmlogerror(pvmtxt);
+		}
+		return -1;
+	}
 
-		len = pvmget32(header + 8);
-		fp = fr_new(len + TDFRAGHDR);
-		bcopy(header, fp->fr_dat, sizeof(header));
-		n = tread(pcbp->tt_fd, fp->fr_dat + TDFRAGHDR, len);
-		if (n != len) {
-			if (n == -1)
-				pvmlogperror("mxinput() tread");
-			else {
-				if (debugmask & TDMPACKET) {
-					sprintf(pvmtxt, "mxinput() t%x read EOF\n", pcbp->tt_tid);
-					pvmlogerror(pvmtxt);
-				}
-			}
+	if (n < 1)
+		return gotem;
+	if ((fp->fr_len += n) < TDFRAGHDR)
+		return gotem;
+
+	/* realloc buffer if frag won't fit */
+
+	m = TDFRAGHDR + pvmget32(fp->fr_dat + 8);
+	if (fp->fr_len == TDFRAGHDR) {
+		if (m > fp->fr_max - (fp->fr_dat - fp->fr_buf)) {
+			fp2 = fr_new(m);
+			BCOPY(fp->fr_dat, fp2->fr_dat, TDFRAGHDR);
+			fp2->fr_len = fp->fr_len;
 			fr_unref(fp);
-			return PvmSysErr;
+			fp = pcbp->tt_rxf = fp2;
+			if (debugmask & TDMPACKET) {
+				sprintf(pvmtxt, "mxinput() realloc frag max=%d\n", m);
+				pvmlogerror(pvmtxt);
+			}
 		}
-		fp->fr_len += TDFRAGHDR + n;
+	}
+
+	if (fp->fr_len == m) {
+		pcbp->tt_rxf = 0;
 
 		cp = fp->fr_dat;
 /*
@@ -950,10 +964,9 @@ mxinput(pcbp)
 			pvmlogerror(pvmtxt);
 		}
 
-		/*
-		* if start of message, make new umbuf, add to frag pile
-		*/
-
+	/*
+	* if start of message, make new umbuf, add to frag pile
+	*/
 		if (ff & FFSOM) {
 			cp += TDFRAGHDR;
 			fp->fr_len -= TTMSGHDR;
@@ -974,15 +987,15 @@ mxinput(pcbp)
 		}
 
 		if (rxup == pcbp->tt_rxfrag) {	/* uh oh, no message for it */
-				pvmlogerror("mxinput() frag with no message\n");
-				fr_unref(fp);
+			pvmlogerror("mxinput() frag with no message\n");
+			fr_unref(fp);
 
 		} else {
 			LISTPUTBEFORE(rxup->ub_frag, fp, fr_link, fr_rlink);
 			rxup->ub_len += fp->fr_len;
-		/*
-		* if end of message, move to pvmrxlist and count it
-		*/
+	/*
+	* if end of message, move to pvmrxlist and count it
+	*/
 			if (ff & FFEOM) {
 				LISTDELETE(rxup, ub_link, ub_rlink);
 				rxup->ub_codef = enctovec(rxup->ub_enc);
@@ -1001,159 +1014,6 @@ mxinput(pcbp)
 	return gotem;
 }
 
-#else	/*!READWHOLEMESSAGE*/
-{
-	int gotem = 0;			/* num msgs added to pvmrxlist */
-	struct frag *rxfp;		/* shadows pcbp->tt_rxf */
-	char *cp;				/* gp */
-	int src;
-	int ff;
-	struct frag *fp;
-	int n;
-	struct umbuf *rxup;
-
-	if (!pcbp->tt_rxf)
-		pcbp->tt_rxf = fr_new(pvmudpmtu);
-	rxfp = pcbp->tt_rxf;
-
-	if (debugmask & TDMPACKET) {
-		sprintf(pvmtxt, "mxinput() pcb t%x fr_len=%d fr_dat=+%d\n",
-				pcbp->tt_tid,
-				rxfp->fr_len, rxfp->fr_dat - rxfp->fr_buf);
-		pvmlogerror(pvmtxt);
-	}
-	if (mxfersingle) {
-		n = (rxfp->fr_len < TDFRAGHDR) ? 0 : pvmget32(rxfp->fr_dat + 8);
-		n += TDFRAGHDR - rxfp->fr_len;
-		n = read(pcbp->tt_fd, rxfp->fr_dat + rxfp->fr_len, n);
-
-	} else {
-		n = read(pcbp->tt_fd,
-				rxfp->fr_dat + rxfp->fr_len,
-				rxfp->fr_max - (rxfp->fr_dat - rxfp->fr_buf) - rxfp->fr_len);
-	}
-
-	if (debugmask & TDMPACKET) {
-		sprintf(pvmtxt, "mxinput() read=%d\n", n);
-		pvmlogerror(pvmtxt);
-	}
-
-	if (n == -1 && errno != EWOULDBLOCK && errno != EINTR) {
-		if (debugmask & TDMPACKET) {
-			sprintf(pvmtxt, "mxinput() t%x read", pcbp->tt_tid);
-			pvmlogperror(pvmtxt);
-		}
-		return PvmSysErr;
-	}
-	if (!n) {
-		if (debugmask & TDMPACKET) {
-			sprintf(pvmtxt, "mxinput() t%x read EOF\n", pcbp->tt_tid);
-			pvmlogerror(pvmtxt);
-		}
-		return PvmSysErr;
-	}
-	if (n > 0) {
-		rxfp->fr_len += n;
-
-		do {
-			fp = rxfp;
-
-			if (fp->fr_len < TDFRAGHDR) {
-/*
-				fprintf(stderr, "mxinput() not a full hdr yet\n");
-*/
-				break;
-			}
-			n = TDFRAGHDR + pvmget32(fp->fr_dat + 8);	/* header + body */
-			if (fp->fr_len < n) {
-/*
-				fprintf(stderr, "mxinput() not %d yet\n", n);
-*/
-				break;
-			}
-
-			if (fp->fr_len > n) {
-				rxfp = pcbp->tt_rxf = fr_new(pvmudpmtu);
-				bcopy(fp->fr_dat + n, rxfp->fr_dat, fp->fr_len - n);
-				rxfp->fr_len = fp->fr_len - n;
-				fp->fr_len = n;
-/*
-				fprintf(stderr, "mxinput() %d left\n", rxfp->fr_len);
-*/
-			} else {
-				rxfp = pcbp->tt_rxf = 0;
-/*
-				fprintf(stderr, "mxinput() zero left\n");
-*/
-			}
-
-			cp = fp->fr_dat;
-/*
-			dst = pvmget32(cp);		XXX might unpack and check
-*/
-			src = pvmget32(cp + 4);
-			ff = pvmget8(cp + 12);
-			fp->fr_len -= TDFRAGHDR;
-			fp->fr_dat += TDFRAGHDR;
-
-			if (debugmask & TDMPACKET) {
-				sprintf(pvmtxt, "mxinput() pkt src t%x len %d ff %d\n",
-						src, fp->fr_len, ff);
-				pvmlogerror(pvmtxt);
-			}
-
-	/*
-	* if start of message, make new umbuf, add to frag pile
-	*/
-			if (ff & FFSOM) {
-				cp += TDFRAGHDR;
-				fp->fr_len -= TTMSGHDR;
-				fp->fr_dat += TTMSGHDR;
-				rxup = midtobuf(umbuf_new());
-				rxup->ub_cod = pvmget32(cp);
-				rxup->ub_enc = pvmget32(cp + 4);
-				rxup->ub_src = src;
-				LISTPUTBEFORE(pcbp->tt_rxfrag, rxup, ub_link, ub_rlink);
-
-			} else {
-				/* locate frag's message */
-
-				for (rxup = pcbp->tt_rxfrag->ub_link; rxup != pcbp->tt_rxfrag;
-						rxup = rxup->ub_link)
-					if (rxup->ub_src == src)
-						break;
-			}
-
-			if (rxup == pcbp->tt_rxfrag) {	/* uh oh, no message for it */
-				pvmlogerror("mxinput() frag with no message\n");
-				fr_unref(fp);
-
-			} else {
-				LISTPUTBEFORE(rxup->ub_frag, fp, fr_link, fr_rlink);
-				rxup->ub_len += fp->fr_len;
-		/*
-		* if end of message, move to pvmrxlist and count it
-		*/
-				if (ff & FFEOM) {
-					LISTDELETE(rxup, ub_link, ub_rlink);
-					rxup->ub_codef = enctovec(rxup->ub_enc);
-					LISTPUTBEFORE(pvmrxlist, rxup, ub_link, ub_rlink);
-					gotem++;
-					if (debugmask & TDMMESSAGE) {
-						sprintf(pvmtxt,
-								"mxinput() src t%x route t%x cod %d len %d\n",
-								rxup->ub_src, pcbp->tt_tid,
-								rxup->ub_cod, rxup->ub_len);
-						pvmlogerror(pvmtxt);
-					}
-				}
-			}
-		} while (rxfp);
-	}
-	return gotem;
-}
-#endif	/*READWHOLEMESSAGE*/
-
 
 /*	mxfer()
 *
@@ -1246,7 +1106,10 @@ int tstkp = 0;
 		if (FD_ISSET(topvmd->tt_fd, &rfds) && !(mxfersingle && gotem)) {
 			up = pvmrxlist->ub_rlink;
 			if ((n = mxinput(topvmd)) < 0) {
-				pvmlogerror("mxfer() mxinput bad return on pvmd sock\n");
+				if (n == -1)
+					pvmlogerror("mxfer() EOF on pvmd sock\n");
+				else
+					pvmlogerror("mxfer() mxinput bad return on pvmd sock\n");
 				return PvmSysErr;
 			}
 
@@ -1344,8 +1207,8 @@ int tstkp = 0;
 				pvmput8(txcp + 12, ff);
 				txtogo += TDFRAGHDR;
 				if (debugmask & TDMPACKET) {
-					sprintf(pvmtxt, "mxfer() dst t%x len %d\n",
-							txup->ub_dst, txfp->fr_len);
+					sprintf(pvmtxt, "mxfer() dst t%x n=%d\n",
+							txup->ub_dst, txtogo);
 					pvmlogerror(pvmtxt);
 				}
 			}
@@ -1471,34 +1334,41 @@ mksocs()
 	int d;
 	int n;
 	int try;
+	char *p;
 
 	if (topvmd)
 		return 0;
 
-	/* get addr of pvmd, make socket to talk */
+	/*
+	* get addr of pvmd, make socket to talk.
+	* first try envar PVMSOCK, then try sockaddr file.
+	*/
 
-	sprintf(buf, TDSOCKNAME, pvm_useruid);
-	if ((d = open(buf, O_RDONLY, 0)) == -1) {
-		pvmlogperror(buf);
-		goto bail;
-	}
-
-	try = 3;
-	do {
-		if ((n = read(d, buf, sizeof(buf))) == -1) {
-			pvmlogperror("mksocs() read addr file");
+	if (!(p = getenv("PVMSOCK"))) {
+		sprintf(buf, TDSOCKNAME, pvm_useruid);
+		if ((d = open(buf, O_RDONLY, 0)) == -1) {
+			pvmlogperror(buf);
 			goto bail;
 		}
-		if (n != 13) {
-			if (--try < 1) {
-				pvmlogerror("mksocs() read addr file: wrong length read\n");
+
+		try = 3;
+		do {
+			if ((n = read(d, buf, sizeof(buf))) == -1) {
+				pvmlogperror("mksocs() read addr file");
 				goto bail;
 			}
-			sleep(1);	/* XXX hmm */
-		}
-	} while (n != 13);
-	(void)close(d);
-	buf[n] = 0;
+			if (n != 13) {
+				if (--try < 1) {
+					pvmlogerror("mksocs() read addr file: wrong length read\n");
+					goto bail;
+				}
+				sleep(1);	/* XXX hmm */
+			}
+		} while (n != 13);
+		(void)close(d);
+		buf[n] = 0;
+		p = buf;
+	}
 
 	FD_ZERO(&pvmrfds);
 /*
@@ -1509,7 +1379,7 @@ mksocs()
 	topvmd = ttpcb_new();
 	topvmd->tt_tid = TIDPVMD;
 
-	hex_inadport(buf, &topvmd->tt_osad);
+	hex_inadport(p, &topvmd->tt_osad);
 	topvmd->tt_osad.sin_family = AF_INET;
 	n = sizeof(topvmd->tt_osad);
 
@@ -1606,7 +1476,7 @@ pvmbeatask()
 	char buf[16];					/* for converting sockaddr */
 
 /*
-	pvm_setdebug(-1);
+	pvm_setopt(PvmDebugMask, -1);
 */
 	if (pvmmytid != -1)
 		return 0;
@@ -1642,11 +1512,11 @@ pvmbeatask()
 	*/
 
 	pvmrxlist = TALLOC(1, struct umbuf, "umb");
-	bzero((char*)pvmrxlist, sizeof(struct umbuf));
+	BZERO((char*)pvmrxlist, sizeof(struct umbuf));
 	pvmrxlist->ub_link = pvmrxlist->ub_rlink = pvmrxlist;
 
 	ttlist = TALLOC(1, struct ttpcb, "tpcb");
-	bzero((char*)ttlist, sizeof(struct ttpcb));
+	BZERO((char*)ttlist, sizeof(struct ttpcb));
 	ttlist->tt_link = ttlist->tt_rlink = ttlist;
 
 	/*
@@ -1737,8 +1607,21 @@ pvmbeatask()
 	}
 	pvm_upkint(&pvmmytid, 1, 1);
 	pvm_upkint(&pvmmyptid, 1, 1);
+
+	pvm_upkint(&pvmouttid, 1, 1);
+	pvm_upkint(&pvmoutcod, 1, 1);
+	pvmcouttid = pvmouttid;
+	pvmcoutcod = pvmoutcod;
+	pvm_upkint(&pvmtrctid, 1, 1);
+	pvm_upkint(&pvmtrccod, 1, 1);
+	pvmctrctid = pvmtrctid;
+	pvmctrccod = pvmtrccod;
+
 	pvm_upkint(&pvmudpmtu, 1, 1);
+	pvmfrgsiz = pvmudpmtu;
+
 	pvm_upkint(&pvmmyndf, 1, 1);
+
 	pvm_upkstr(buf);
 	hex_inadport(buf, &pvmourinet);
 	pvmourinet.sin_family = AF_INET;
@@ -1787,21 +1670,143 @@ pvmendtask()
 
 
 int
-pvm_advise(what)
+pvm_getopt(what)
 	int what;
 {
+	int rc = 0;
+	int err = 0;
+
 	switch (what) {
-	case PvmDontRoute:
-	case PvmAllowDirect:
-	case PvmRouteDirect:
-		pvmrouteopt = what;
+	case PvmRoute:
+		rc = pvmrouteopt;
+		break;
+
+	case PvmDebugMask:
+		rc = debugmask;
+		break;
+
+	case PvmAutoErr:
+		rc = pvmautoerr;
+		break;
+
+	case PvmOutputTid:
+		rc = pvmcouttid;
+		break;
+
+	case PvmOutputCode:
+		rc = pvmcoutcod;
+		break;
+
+	case PvmTraceTid:
+		rc = pvmctrctid;
+		break;
+
+	case PvmTraceCode:
+		rc = pvmctrccod;
+		break;
+
+	case PvmFragSize:
+		rc = pvmfrgsiz;
 		break;
 
 	default:
-		return lpvmerr("pvm_advise", PvmBadParam);
+		err = 1;
 		break;
 	}
-	return 0;
+	if (err)
+		return lpvmerr("pvm_getopt", PvmBadParam);
+	return rc;
+}
+
+
+int
+pvm_setopt(what, val)
+	int what;
+	int val;
+{
+	int rc = 0;
+	int err = 0;
+
+	switch (what) {
+	case PvmRoute:
+		switch (val) {
+		case PvmDontRoute:
+		case PvmAllowDirect:
+		case PvmRouteDirect:
+			rc = pvmrouteopt;
+			pvmrouteopt = val;
+			break;
+
+		default:
+			err = 1;
+			break;
+		}
+		break;
+
+	case PvmDebugMask:
+		rc = debugmask;
+		debugmask = val;
+		break;
+
+	case PvmAutoErr:
+		rc = pvmautoerr;
+		pvmautoerr = val;
+		break;
+
+	case PvmOutputTid:
+		if (val && val != pvmmytid
+		&& (val != pvmouttid || pvmcoutcod != pvmoutcod))
+			err = 1;
+		else {
+			rc = pvmcouttid;
+			pvmcouttid = val;
+		}
+		break;
+
+	case PvmOutputCode:
+		if (pvmcouttid && pvmcouttid != pvmmytid && val != pvmoutcod)
+			err = 1;
+		else {
+			rc = pvmcoutcod;
+			pvmcoutcod = val;
+		}
+		break;
+
+	case PvmTraceTid:
+		if (val && val != pvmmytid
+		&& (val != pvmtrctid || pvmctrccod != pvmtrccod))
+			err = 1;
+		else {
+			rc = pvmctrctid;
+			pvmctrctid = val;
+		}
+		break;
+
+	case PvmTraceCode:
+		if (pvmctrctid && pvmctrctid != pvmmytid && val != pvmtrccod)
+			err = 1;
+		else {
+			rc = pvmctrccod;
+			pvmctrccod = val;
+		}
+		break;
+
+	case PvmFragSize:
+		if (val < TDFRAGHDR + TTMSGHDR + 4 || val > 1048576)
+			err = 1;
+		else {
+			rc = pvmfrgsiz;
+			pvmfrgsiz = val;
+		}
+		break;
+
+	default:
+		err = 1;
+		break;
+	}
+	if (err)
+		return lpvmerr("pvm_setopt", PvmBadParam);
+	return rc;
 }
 
 
@@ -1821,30 +1826,6 @@ pvm_perror(s)
 }
 
 
-int
-pvm_setdebug(mask)
-	int mask;
-{
-	int old = debugmask;
-
-	debugmask = mask;
-	return old;
-}
-
-
-#if 0
-int
-pvm_getfd()
-{
-	int cc;
-
-	if (cc = pvmbeatask())
-		return lpvmerr("pvm_getfd", cc);
-	return topvmd ? topvmd->tt_fd : -1;
-}
-#endif
-
-
 int
 pvm_getfds(fds)		/* XXX this function kinda sucks */
 	int **fds;			/* fd list return */
@@ -1882,15 +1863,19 @@ pvm_getfds(fds)		/* XXX this function kinda sucks */
 
 
 int
-pvm_start_pvmd(argc, argv)
-	int argc;
-	char **argv;
+pvm_start_pvmd(argc, argv, block)
+	int argc;		/* number of args to pass to pvmd (>= 0) */
+	char **argv;	/* args for pvmd or null */
+	int block;		/* if true, don't return until add hosts are started */
 {
 	char buf[128];
 	struct stat sb;
 	int cc;
 	char *fn;			/* file to exec */
-	char *hd;			/* homedir */
+	char **av;
+
+	if (argc < 0 || !argv)
+		argc = 0;
 
 	if ((pvm_useruid = getuid()) == -1) {
 		pvmlogerror("can't getuid()\n");
@@ -1902,24 +1887,14 @@ pvm_start_pvmd(argc, argv)
 	if (stat(buf, &sb) != -1)
 		return lpvmerr("pvm_start_pvmd", PvmDupHost);
 
-	if (PVMDFILE[0] == '/') {
-		hd = "";
+	fn = pvmgetpvmd();
 
-	} else {
-		if (!(hd = getenv("HOME"))) {
-			pvmlogerror("can't getenv(HOME)\n");
-			return lpvmerr("pvm_start_pvmd", PvmSysErr);
-		}
-	}
+	av = TALLOC(argc + 2, char *, "argv");
+	if (argc > 0)
+		BCOPY(&argv[0], &av[1], argc * sizeof(char*));
+	av[0] = fn;
+	av[argc + 1] = 0;
 
-	if (!(fn = TALLOC(strlen(hd) + strlen(PVMDFILE) + 2, char, "XXX"))) {
-		pvmlogerror("can't malloc\n");
-		return lpvmerr("pvm_start_pvmd", PvmSysErr);
-	}
-	sprintf(fn, "%s/%s", hd, PVMDFILE);
-
-	argv[0] = fn;
-	argv[argc] = 0;
 	if (!fork()) {
 		if (!fork()) {
 			for (cc = getdtablesize(); cc-- > 0; )
@@ -1927,22 +1902,62 @@ pvm_start_pvmd(argc, argv)
 			(void)open("/dev/null", O_RDONLY, 0);
 			(void)open("/dev/null", O_WRONLY, 0);
 			(void)dup2(1, 2);
-			execvp(argv[0], argv);
+			execvp(av[0], av);
 		}
 		_exit(0);
 	}
 	(void)wait(0);
 
+	PVM_FREE(av);
 	PVM_FREE(fn);
 
 	for (cc = 8; cc > 0 && stat(buf, &sb) == -1; cc--)
 		sleep(1);
 
-	if (cc > 0)
-		cc = pvm_mytid();
-	else
+	if (cc <= 0) {
 		cc = PvmCantStart;
+		goto bail;
+	}
+	if ((cc = pvm_mytid()) < 0)
+		goto bail;
+
+	if (block) {
+		struct hostinfo *hip;
+		int t = 1;
+
+		pvm_config((int*)0, (int*)0, &hip);
+		while ((cc = pvm_addhosts(&hip[0].hi_name, 1, (int*)0)) == PvmAlready) {
+			sleep(t);
+			if (t < 8)
+				t *= 2;
+		}
+		if (cc != PvmDupHost)
+			goto bail;
+		cc = pvm_mytid();
+	}
+
+bail:
 	return (cc < 0 ? lpvmerr("pvm_start_pvmd", cc) : 0);
 }
 
 
+/*
+* Obsolete Functions
+*/
+
+int
+pvm_advise(what)
+	int what;
+{
+	return pvm_setopt(PvmRoute, what);
+}
+
+
+int
+pvm_setdebug(mask)
+	int mask;
+{
+	return pvm_setopt(PvmDebugMask, mask);
+}
+
+
