@@ -1,11 +1,11 @@
 
 /*
- *           PVM 3.0:  Parallel Virtual Machine System 3.0
+ *           PVM 3.2:  Parallel Virtual Machine System 3.2
  *               University of Tennessee, Knoxville TN.
  *           Oak Ridge National Laboratory, Oak Ridge TN.
  *                   Emory University, Atlanta GA.
  *      Authors:  A. L. Beguelin, J. J. Dongarra, G. A. Geist,
- *          R. J. Manchek, B. K. Moore, and V. S. Sunderam
+ *    W. C. Jiang, R. J. Manchek, B. K. Moore, and V. S. Sunderam
  *                   (C) 1992 All Rights Reserved
  *
  *                              NOTICE
@@ -22,7 +22,7 @@
  * purpose.  This software is provided ``as is'' without express or
  * implied warranty.
  *
- * PVM 3.0 was funded in part by the U.S. Department of Energy, the
+ * PVM 3.2 was funded in part by the U.S. Department of Energy, the
  * National Science Foundation and the State of Tennessee.
  */
 
@@ -31,7 +31,10 @@
  *
  *	Libpvm generic functions.
  *
-$Log$
+$Log: lpvmgen.c,v $
+ * Revision 1.1  1993/08/30  23:26:48  manchek
+ * Initial revision
+ *
  */
 
 
@@ -39,20 +42,26 @@ $Log$
 #ifdef IMA_BSD386
 #include <machine/endian.h>
 #endif
+#ifdef IMA_LINUX
+#include <endian.h>
+#endif
 #include <rpc/types.h>
 #include <rpc/xdr.h>
 #ifdef	SYSVSTR
 #include <string.h>
+#define	CINDEX(s,c)	strchr(s,c)
 #else
 #include <strings.h>
+#define	CINDEX(s,c)	index(s,c)
 #endif
 #include <signal.h>
-#include "pvm3.h"
+#include <pvm3.h>
 #include "global.h"
 #include "pvmalloc.h"
 #include "tdpro.h"
 #include "pvmumbuf.h"
 #include "listmac.h"
+#include "bfunc.h"
 
 
 /***************
@@ -60,7 +69,14 @@ $Log$
  **           **
  ***************/
 
+char *getenv();
+
 extern int pvm_errno;				/* from lpvm.c */
+extern int pvmautoerr;				/* from lpvm.c */
+extern int pvmcouttid;				/* from lpvm.c */
+extern int pvmcoutcod;				/* from lpvm.c */
+extern int pvmctrctid;				/* from lpvm.c */
+extern int pvmctrccod;				/* from lpvm.c */
 extern int pvmmyptid;				/* from lpvm.c */
 extern int pvmmytid;				/* from lpvm.c */
 extern int pvmrbufmid;				/* from pack.c */
@@ -76,7 +92,7 @@ extern int pvmtidhmask;				/* from lpvm.c */
 
 static int def_match();
 
-static int pvmautoerr = 1;				/* whether to auto print err msgs */
+static char rcsid[] = "$Id: lpvmgen.c,v 1.1 1993/08/30 23:26:48 manchek Exp $";
 static char pvmtxt[512];				/* scratch for error log */
 static int (*recv_match)() = def_match;
 
@@ -163,7 +179,7 @@ pvm_addhosts(names, count, svp)
 						if (sv[i] >= 0)
 							cc++;
 					if (svp)
-						bcopy((char*)sv, (char*)svp, count * sizeof(int));
+						BCOPY((char*)sv, (char*)svp, count * sizeof(int));
 					PVM_FREE(sv);
 
 				} else {
@@ -220,7 +236,6 @@ pvm_config(nhostp, narchp, hostp)
 				hlist[i].hi_name = STRALLOC(buf);
 				pvm_upkstr(buf);
 				hlist[i].hi_arch = STRALLOC(buf);
-				pvm_upkint(&hlist[i].hi_mtu, 1, 1);
 				pvm_upkint(&hlist[i].hi_speed, 1, 1);
 			}
 			pvm_freebuf(pvm_setrbuf(rbf));
@@ -304,7 +319,7 @@ pvm_delhosts(names, count, svp)
 						if (sv[i] >= 0)
 							cc++;
 					if (svp)
-						bcopy((char*)sv, (char*)svp, count * sizeof(int));
+						BCOPY((char*)sv, (char*)svp, count * sizeof(int));
 					PVM_FREE(sv);
 
 				} else {
@@ -360,11 +375,9 @@ pvm_halt()
 	if (!(cc = pvmbeatask())) {
 		sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
 		rbf = pvm_setrbuf(0);
-		if ((cc = msendrecv(TIDPVMD, TM_HALT)) > 0)
-			pvm_freebuf(pvm_setrbuf(rbf));
-		else
-			pvm_setrbuf(rbf);
+		cc = (msendrecv(TIDPVMD, TM_HALT) < 0) ? 0 : PvmSysErr;
 		pvm_freebuf(pvm_setsbuf(sbf));
+		pvm_setrbuf(rbf);
 	}
 	return (cc < 0 ? lpvmerr("pvm_halt", cc) : 0);
 }
@@ -410,7 +423,7 @@ int
 pvm_kill(tid)
 	int tid;
 {
-	return pvm_sendsig(tid, SIGKILL);
+	return pvm_sendsig(tid, SIGTERM);
 }
 
 
@@ -467,8 +480,10 @@ pvm_mcast(tids, count, code)
 		return lpvmerr("pvm_mcast", cc);
 	if (pvmsbufmid <= 0)
 		return lpvmerr("pvm_mcast", PvmNoBuf);
-	if ((code & ~0x7fffffff) || count < 1)
+	if ((code & ~0x7fffffff) || count < 0)
 		return lpvmerr("pvm_mcast", PvmBadParam);
+	if (!count)
+		return 0;
 	for (i = count; i-- > 0; )
 		if (!TIDISTASK(tids[i]))
 			return lpvmerr("pvm_mcast", PvmBadParam);
@@ -492,9 +507,6 @@ pvm_mcast(tids, count, code)
 
 	/* send message */
 
-/* XXX short-ckt to us should go here.  maybe can inc frag chain
-   XXX count and make new message, put on pvmrxlist. */
-
 	if ((cc = mroute(pvmsbufmid, mca, code, 0)) < 0)
 		lpvmerr("pvm_mcast", cc);
 	return 0;
@@ -571,7 +583,11 @@ pvm_notify(what, code, count, vals)
 		break;
 
 	case PvmHostAdd:
+/*
 		return lpvmerr("pvm_notify", PvmNotImpl);
+*/
+count = 0;
+vals = &count;
 		break;
 
 	default:
@@ -591,64 +607,6 @@ pvm_notify(what, code, count, vals)
 	return (cc < 0 ? lpvmerr("pvm_notify", cc) : 0);
 }
 
-#if 0
-int
-pvm_notify(what, code, count, vals)
-	int what;
-	int code;
-	int count;
-	int *vals;
-{
-	int sbf, rbf;
-	int cc;
-
-	if (cc = pvmbeatask())
-		return lpvmerr("pvm_notify", cc);
-
-	if (code & ~0x7fffffff)
-		return lpvmerr("pvm_notify", PvmBadParam);
-
-	switch (what) {
-
-	case PvmHostDelete:
-		if (count < 1)
-			return lpvmerr("pvm_notify", PvmBadParam);
-		break;
-
-	case PvmTaskExit:
-		if (count < 1)
-			return lpvmerr("pvm_notify", PvmBadParam);
-		for (cc = count; cc-- > 0; )
-			if (!TIDISTASK(vals[cc]))
-				return lpvmerr("pvm_notify", PvmBadParam);
-		break;
-
-	case PvmHostAdd:
-		return lpvmerr("pvm_notify", PvmNotImpl);
-		break;
-
-	default:
-		return lpvmerr("pvm_notify", PvmBadParam);
-		break;
-	}
-
-	sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
-	rbf = pvm_setrbuf(0);
-	pvm_pkint(&what, 1, 1);
-	pvm_pkint(&code, 1, 1);
-	pvm_pkint(&count, 1, 1);
-	pvm_pkint(vals, count, 1);
-
-	if ((cc = msendrecv(TIDPVMD, TM_NOTIFY)) > 0)
-		pvm_freebuf(pvm_setrbuf(rbf));
-	else
-		pvm_setrbuf(rbf);
-	pvm_freebuf(pvm_setsbuf(sbf));
-
-	return (cc < 0 ? lpvmerr("pvm_notify", cc) : 0);
-}
-#endif
-
 
 int
 pvm_nrecv(tid, code)
@@ -694,6 +652,7 @@ pvm_nrecv(tid, code)
 	}
 
 	LISTDELETE(bestup, ub_link, ub_rlink);
+	bestup->ub_flag &= ~(UB_PACK|UB_UPACK);
 	if (cc = pvm_setrbuf(bestup->ub_mid))
 		return cc;
 	return bestup->ub_mid;
@@ -748,6 +707,7 @@ pvm_probe(tid, code)
 			bestup = up;
 		}
 	}
+	bestup->ub_flag &= ~(UB_PACK|UB_UPACK);
 	return bestup->ub_mid;
 }
 
@@ -820,6 +780,7 @@ pvm_recv(tid, code)
 	}
 
 	LISTDELETE(bestup, ub_link, ub_rlink);
+	bestup->ub_flag &= ~(UB_PACK|UB_UPACK);
 	if (cc = pvm_setrbuf(bestup->ub_mid))
 		return cc;
 	return bestup->ub_mid;
@@ -868,6 +829,9 @@ pvm_sendsig(tid, signum)
 	int sbf, rbf;
 
 	if (!(cc = pvmbeatask())) {
+		if (!TIDISTASK(tid))
+			return lpvmerr("pvm_sendsig", PvmBadParam);
+
 		sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
 		rbf = pvm_setrbuf(0);
 
@@ -883,17 +847,6 @@ pvm_sendsig(tid, signum)
 }
 
 
-int
-pvm_serror(how)
-	int how;
-{
-	int old = pvmautoerr;
-
-	pvmautoerr = how;
-	return old;
-}
-
-
 /*	bubble()
 *
 *	Move nonnegative els to head of array, negative ones to end.
@@ -924,6 +877,58 @@ bubble(n, a)
 }
 
 
+#if	defined(IMA_PGON) || defined(IMA_I860) || defined(IMA_CM5)
+static int
+pvmgetenvars(ep)
+	char ***ep;
+{
+	return 0;
+}
+
+#else	/*defined(IMA_PGON) || defined(IMA_I860) || defined(IMA_CM5)*/
+static int
+pvmgetenvars(ep)
+	char ***ep;
+{
+	char **xpl;			/* vars to export */
+	int mxpl;			/* cur length of xpl */
+	int nxpl;			/* num vars found */
+	char buf[200];
+	char *p, *q;
+	int n;
+
+	if (p = getenv("PVM_EXPORT")) {
+		mxpl = 5;
+		xpl = TALLOC(mxpl, char *, "env");
+		xpl[0] = p - 11;
+		nxpl = 1;
+		while (1) {
+			while (*p == ':')
+				p++;
+			if (!*p)
+				break;
+			n = (q = CINDEX(p, ':')) ? q - p : strlen(p);
+			strncpy(buf, p, n);
+			buf[n] = 0;
+			if (q = getenv(buf)) {
+				if (nxpl == mxpl) {
+					mxpl += mxpl / 2 + 1;
+					xpl = TREALLOC(xpl, mxpl, char *);
+				}
+				xpl[nxpl++] = q - n - 1;
+			}
+			p += n;
+		}
+		*ep = xpl;
+		return nxpl;
+
+	} else {
+		return 0;
+	}
+}
+#endif	/*defined(IMA_PGON) || defined(IMA_I860) || defined(IMA_CM5)*/
+
+
 int
 pvm_spawn(file, argv, flags, where, count, tids)
 	char *file;
@@ -936,6 +941,7 @@ pvm_spawn(file, argv, flags, where, count, tids)
 	int sbf, rbf;	/* temp for current tx, rx msgs */
 	int cc;
 	int i, n;
+	char **ep;
 
 	if (!(cc = pvmbeatask())) {
 		if (count < 1)
@@ -956,6 +962,19 @@ pvm_spawn(file, argv, flags, where, count, tids)
 		for (i = 0; i < n; i++)
 			pvm_pkstr(argv[i]);
 
+		pvm_pkint(&pvmcouttid, 1, 1);
+		pvm_pkint(&pvmcoutcod, 1, 1);
+		pvm_pkint(&pvmctrctid, 1, 1);
+		pvm_pkint(&pvmctrccod, 1, 1);
+
+		if ((n = pvmgetenvars(&ep)) > 0) {
+			pvm_pkint(&n, 1, 1);
+			for (i = 0; i < n; i++)
+				pvm_pkstr(ep[i]);
+			PVM_FREE(ep);
+		} else
+			pvm_pkint(&n, 1, 1);
+
 		if ((cc = msendrecv(TIDPVMD, TM_SPAWN)) > 0) {
 			pvm_upkint(&cc, 1, 1);
 			if (tids) {
@@ -1010,7 +1029,7 @@ pvm_tasks(where, ntaskp, taskp)
 			if (!(cc = pvm_upkint(&ec, 1, 1))
 			&& (cc = ec) >= 0) {
 				tlist = TALLOC(len1, struct taskinfo, "ti");
-				ae = pvm_serror(0);
+				ae = pvm_setopt(PvmAutoErr, 0);
 				ntask = 0;
 				while (!pvm_upkint(&tlist[ntask].ti_tid, 1, 1)) {
 					pvm_upkint(&tlist[ntask].ti_ptid, 1, 1);
@@ -1025,7 +1044,7 @@ pvm_tasks(where, ntaskp, taskp)
 						tlist = TREALLOC(tlist, len1, struct taskinfo);
 					}
 				}
-				pvm_serror(ae);
+				pvm_setopt(PvmAutoErr, ae);
 			}
 			pvm_freebuf(pvm_setrbuf(rbf));
 			if (ntaskp)
@@ -1074,26 +1093,6 @@ pvm_tickle(narg, argp, nresp, resp)
 	return (cc < 0 ? lpvmerr("pvm_tickle", cc) : 0);
 }
 
-#if 0
-int
-pvm_tickle(how)
-	int how;
-{
-	int cc, sbf, rbf;
-
-	if (!(cc = pvmbeatask())) {
-		sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
-		rbf = pvm_setrbuf(0);
-		pvm_pkint(&how, 1, 1);
-		if ((cc = msendrecv(TIDPVMD, TM_TICKLE)) > 0)
-			pvm_freebuf(pvm_setrbuf(rbf));
-		else
-			pvm_setrbuf(rbf);
-		pvm_freebuf(pvm_setsbuf(sbf));
-	}
-	return (cc < 0 ? lpvmerr("pvm_tickle", cc) : 0);
-}
-#endif
 
 int
 pvm_tidtohost(tid)
@@ -1110,3 +1109,15 @@ pvm_version()
 }
 
 
+/*
+* Obsolete Functions
+*/
+
+int
+pvm_serror(how)
+	int how;
+{
+	return pvm_setopt(PvmAutoErr, how);
+}
+
+
