@@ -1,11 +1,11 @@
 
 /*
- *           PVM 3.0:  Parallel Virtual Machine System 3.0
+ *           PVM 3.2:  Parallel Virtual Machine System 3.2
  *               University of Tennessee, Knoxville TN.
  *           Oak Ridge National Laboratory, Oak Ridge TN.
  *                   Emory University, Atlanta GA.
  *      Authors:  A. L. Beguelin, J. J. Dongarra, G. A. Geist,
- *          R. J. Manchek, B. K. Moore, and V. S. Sunderam
+ *    W. C. Jiang, R. J. Manchek, B. K. Moore, and V. S. Sunderam
  *                   (C) 1992 All Rights Reserved
  *
  *                              NOTICE
@@ -22,7 +22,7 @@
  * purpose.  This software is provided ``as is'' without express or
  * implied warranty.
  *
- * PVM 3.0 was funded in part by the U.S. Department of Energy, the
+ * PVM 3.2 was funded in part by the U.S. Department of Energy, the
  * National Science Foundation and the State of Tennessee.
  */
 
@@ -31,7 +31,10 @@
  *
  *	Entry points for messages from local tasks.
  *
-$Log$
+$Log: tdpro.c,v $
+ * Revision 1.1  1993/08/30  23:26:52  manchek
+ * Initial revision
+ *
  */
 
 
@@ -59,6 +62,7 @@ $Log$
 #include "host.h"
 #include "waitc.h"
 #include "listmac.h"
+#include "bfunc.h"
 
 #ifndef	min
 #define	min(a,b)	((a)<(b)?(a):(b))
@@ -79,6 +83,7 @@ extern struct task *locltasks;		/* from task.c */
 extern int myndf;					/* from pvmd.c */
 extern int myhostpart;				/* from pvmd.c */
 extern int mytid;					/* from pvmd.c */
+extern int nopax;					/* from pvmd.c */
 extern int ourudpmtu;				/* from pvmd.c */
 extern int runstate;				/* from pvmd.c */
 extern int tidhmask;				/* from pvmd.c */
@@ -90,7 +95,7 @@ extern int tidlmask;				/* from pvmd.c */
  **           **
  ***************/
 
-static char rcsid[] = "$Id$";
+static char rcsid[] = "$Id: tdpro.c,v 1.1 1993/08/30 23:26:52 manchek Exp $";
 static char pvmtxt[512];		/* scratch for error log */
 
 int tm_addhost();
@@ -211,13 +216,6 @@ bail:
  *********************************/
 
 
-/*	tm_connect()
-*
-*	First message from a task.
-*	We write in t-auth file to prove our ident,
-*	then make d-auth file for task to prove itself.
-*/
-
 #ifdef	NOTMPNAM
 #define	LEN_OF_TMP_NAM	32
 char *pvmtmpnam();
@@ -231,6 +229,24 @@ char *pvmtmpnam();
 
 #endif	/*NOTMPNAM*/
 
+/*	tm_connect()
+*
+*	Task connecting to pvmd phase 1.
+*	We assign it a context, write in t-auth file to prove our ident,
+*	then make d-auth file for task to prove itself.
+*
+*	TM_CONNECT
+*	call {
+*		int tdprotocol		// t-d protocol number compiled into libpvm
+*		string authfile		// t-auth file pvmd must write
+*	}
+*	ret {
+*		int tdprotocol		// pvmd's td protocol version
+*		int acknack			// 1 if pvmd accepts connection
+*		string authfile		// d-auth file task must write
+*	}
+*/
+
 int
 tm_connect(tp, mp)
 	struct task *tp;
@@ -315,8 +331,25 @@ bail:
 
 /*	tm_conn2()
 *
-*	Second message from a task.
-*	We check d-auth file and give it config info.
+*	Task connecting to pvmd phase 2.
+*	We check d-auth file and give it config info, attach to real context.
+*
+*	TM_CONN2
+*	call {
+*		int unixpid			// real pid
+*		int identpid		// pid assigned by pvmd or 0
+*	}
+*	ret {
+*		int acknack			// 1 if pvmd accepts connection
+*		int tid				// task tid
+*		int ptid			// parent tid
+*		int outtid			// output dst and code
+*		int outcod
+*		int trctid			// trace dst and code
+*		int trccod
+*		int udpmax
+*		int nativecode
+*	}
 */
 
 int
@@ -383,7 +416,7 @@ tm_conn2(tp, mp)
 		}
 		tp2 = task_new(tid);
 		if (debugmask & PDMTASK) {
-			sprintf(pvmtxt, "tm_connect() new task t%x\n", tp2->t_tid);
+			sprintf(pvmtxt, "tm_conn2() new task t%x\n", tp2->t_tid);
 			pvmlogerror(pvmtxt);
 		}
 	}
@@ -413,6 +446,10 @@ tm_conn2(tp, mp)
 	pkint(mp, 1);
 	pkint(mp, tp->t_tid);
 	pkint(mp, tp->t_ptid);
+	pkint(mp, tp->t_outtid);
+	pkint(mp, tp->t_outcod);
+	pkint(mp, tp->t_trctid);
+	pkint(mp, tp->t_trccod);
 	pkint(mp, ourudpmtu);
 	pkint(mp, myndf);
 	pkstr(mp, inadport_hex(&(hosts->ht_hosts[hosts->ht_local]->hd_sad)));
@@ -431,6 +468,10 @@ bail:
 /*	tm_exit()
 *
 *	Last message from a task.  This is the nice way of disconnecting.
+*
+*	TM_EXIT
+*	call { }
+*	ret { }
 */
 
 int
@@ -441,13 +482,26 @@ tm_exit(tp, mp)
 	mp = mesg_new(0);
 	mp->m_dst = tp->t_tid;
 	mp->m_cod = TM_EXIT;
+	tp->t_flag |= TF_CLOSE;
 	sendmessage(mp);
 	task_cleanup(tp);
-	tp->t_flag |= TF_CLOSE;
 	return 0;
 }
 
 
+/*	tm_pstat()
+*
+*	Task wants status of another task.
+*
+*	TM_PSTAT
+*	call {
+*		int tid
+*	}
+*	ret {
+*		int status
+*	}
+*/
+
 int
 tm_pstat(tp, mp)
 	struct task *tp;
@@ -500,6 +554,16 @@ tm_pstat(tp, mp)
 *
 *	Task requesting to add hosts to virtual machine.  Exit point is
 *	here or dm_addhostack().
+*
+*	TM_ADDHOST
+*	call {
+*		int nhosts
+*		string names[nhosts]
+*	}
+*	ret {
+*		int nhosts			// or negative for error
+*		int status[nhosts]	// status (tid if successful) of each host
+*	}
 */
 
 int
@@ -556,6 +620,12 @@ free_wait_spawn(wxp)
 				PVM_FREE(wxp->w_argv[i]);
 		PVM_FREE(wxp->w_argv);
 	}
+	if (wxp->w_env) {
+		for (i = 0; i < wxp->w_nenv; i++)
+			if (wxp->w_env[i])
+				PVM_FREE(wxp->w_env[i]);
+		PVM_FREE(wxp->w_env);
+	}
 	if (wxp->w_ht)
 		ht_free(wxp->w_ht);
 	if (wxp->w_vec)
@@ -569,6 +639,26 @@ free_wait_spawn(wxp)
 *
 *	Task requesting to spawn other tasks.  Exit point for this
 *	request is here or dm_spawnack().
+*
+*	TM_SPAWN
+*	call {
+*		string file
+*		int flags
+*		string where
+*		int count
+*		int nargs
+*		string argv[nargs]
+*		int outtid
+*		int outcod
+*		int trctid
+*		int trccod
+*		int nenv
+*		string env[nenv]
+*	}
+*	ret {
+*		int ntids  // or nack
+*		int tids[acknack]
+*	}
 */
 
 int
@@ -582,6 +672,10 @@ tm_spawn(tp, mp)
 	int count;					/* num of tasks */
 	int argc;					/* num of args */
 	char **argv = 0;			/* args */
+	int outtid, outcod;			/* stdout dst and code */
+	int trctid, trccod;			/* trace dst and code */
+	int nenv;					/* num of envars */
+	char **env = 0;				/* envars */
 
 	struct waitc *wp;			/* 'seed' waitc */
 	struct waitc_spawn *wxp;	/* extra data */
@@ -593,6 +687,7 @@ tm_spawn(tp, mp)
 
 	/*
 	* unpack spawn command from task
+	* XXX should clean up this code and upk directly into waitc_spawn
 	*/
 
 	if (upkstralloc(mp, &file)
@@ -606,13 +701,25 @@ tm_spawn(tp, mp)
 		goto bad;
 
 	argc++;
-	bzero((char*)(argv = TALLOC(argc, char*, "argv")),
+	BZERO((char*)(argv = TALLOC(argc, char*, "argv")),
 			argc * sizeof(char*));
 	argc--;
 	for (i = 0; i < argc; i++)
 		if (upkstralloc(mp, &argv[i]))
 			goto bad;
 
+	if (upkuint(mp, &outtid)
+	|| upkuint(mp, &outcod)
+	|| upkuint(mp, &trctid)
+	|| upkuint(mp, &trccod)
+	|| upkuint(mp, &nenv))
+		goto bad;
+	BZERO((char*)(env = TALLOC((nenv + 1), char*, "env")),
+			(nenv + 1) * sizeof(char*));
+	for (i = 0; i < nenv; i++)
+		if (upkstralloc(mp, &env[i]))
+			goto bad;
+
 	/*
 	* make host set containing hosts (matching where option)
 	*/
@@ -636,9 +743,21 @@ tm_spawn(tp, mp)
 			ht_merge(htp, hosts);
 		}
 	}
-/*
-	ht_dump(htp);
-*/
+
+	if (flags & PvmHostCompl) {
+		for (hh = hosts->ht_last; hh > 0; hh--) {
+			if (hh <= htp->ht_last && (hp = htp->ht_hosts[hh]))
+				ht_delete(htp, hp);
+			else
+				if (hp = hosts->ht_hosts[hh])
+					ht_insert(htp, hp);
+		}
+	}
+
+	if (debugmask & PDMTASK) {
+		pvmlogerror("tm_spawn() host set:\n");
+		ht_dump(htp);
+	}
 
 	/*
 	* assign each task to a host
@@ -648,7 +767,7 @@ tm_spawn(tp, mp)
 	wp->wa_tid = tp->t_tid;
 
 	wxp = TALLOC(1, struct waitc_spawn, "waix");
-	bzero((char*)wxp, sizeof(struct waitc_spawn));
+	BZERO((char*)wxp, sizeof(struct waitc_spawn));
 	wp->wa_spec = (void*)wxp;
 
 	wxp->w_file = file;
@@ -657,9 +776,16 @@ tm_spawn(tp, mp)
 	wxp->w_argc = argc;
 	wxp->w_argv = argv;
 	argv = 0;
+	wxp->w_outtid = outtid;
+	wxp->w_outcod = outcod;
+	wxp->w_trctid = trctid;
+	wxp->w_trccod = trccod;
+	wxp->w_nenv = nenv;
+	wxp->w_env = env;
+	env = 0;
 	wxp->w_ht = htp;
 	vec = TALLOC(count, int, "vec");
-	bzero((char*)vec, count * sizeof(int));
+	BZERO((char*)vec, count * sizeof(int));
 	wxp->w_vec = vec;
 	wxp->w_togo = wxp->w_veclen = count;
 
@@ -689,6 +815,12 @@ cleanup:
 				PVM_FREE(argv[i]);
 		PVM_FREE(argv);
 	}
+	if (env) {
+		for (i = 0; i < nenv; i++)
+			if (env[i])
+				PVM_FREE(env[i]);
+		PVM_FREE(env);
+	}
 	return 0;
 }
 
@@ -712,11 +844,12 @@ assign_tasks(wp)
 	int na = 0;						/* num tasks assigned to a host */
 	struct waitc *wp2;
 	struct hostd *hp;
-	struct mesg *mp;
+	struct mesg *mp, *mp2;
 	int t;
 	int i;
+	int tid;
 
-	static int lasthh = 0;		/* for assigning hosts */
+	static int lasthh = -1;		/* for assigning hosts */
 
 	if (!wxp)
 		return 0;
@@ -747,8 +880,19 @@ assign_tasks(wp)
 		mp->m_dst = wp->wa_tid;
 		mp->m_cod = TM_SPAWN;
 		pkint(mp, wxp->w_veclen);
-		for (t = 0; t < wxp->w_veclen; t++)
-			pkint(mp, wxp->w_vec[t]);
+		for (t = 0; t < wxp->w_veclen; t++) {
+			tid = wxp->w_vec[t];
+			pkint(mp, tid);
+			if (TIDISTASK(tid) && wxp->w_outtid) {
+				mp2 = mesg_new(0);
+				mp2->m_cod = wxp->w_outcod;
+				mp2->m_dst = wxp->w_outtid;
+				pkint(mp2, tid);
+				pkint(mp2, -1);
+				pkint(mp2, wp->wa_tid);
+				sendmessage(mp2);
+			}
+		}
 		sendmessage(mp);
 
 		free_wait_spawn(wxp);
@@ -764,6 +908,8 @@ assign_tasks(wp)
 
 	/* find first host to assign */
 
+	if (lasthh == -1)
+		lasthh = hosts->ht_local + 1;
 	if (lasthh > htp->ht_last)
 		lasthh = 0;
 	while (!htp->ht_hosts[lasthh])
@@ -801,6 +947,13 @@ assign_tasks(wp)
 			pkint(mp, wxp->w_argc);
 			for (i = 0; i < wxp->w_argc; i++)
 				pkstr(mp, wxp->w_argv[i]);
+			pkint(mp, wxp->w_outtid);
+			pkint(mp, wxp->w_outcod);
+			pkint(mp, wxp->w_trctid);
+			pkint(mp, wxp->w_trccod);
+			pkint(mp, wxp->w_nenv);
+			for (i = 0; i < wxp->w_nenv; i++)
+				pkstr(mp, wxp->w_env[i]);
 			mp->m_dst = hp->hd_hostpart | TIDPVMD;
 			mp->m_cod = DM_SPAWN;
 			mp->m_wid = wp2->wa_wid;
@@ -824,6 +977,13 @@ assign_tasks(wp)
 /*	tm_sendsig()
 *
 *	Task sending a signal to another task.
+*
+*	TM_SENDSIG
+*	call {
+*		int tid
+*		int signum
+*	}
+*	ret { }
 */
 
 int
@@ -858,7 +1018,21 @@ tm_sendsig(tp, mp)
 
 /*	tm_config()
 *
-*	Task wants machine configuration
+*	Task wants machine configuration.
+*
+*	TM_CONFIG
+*	call { }
+*	ret {
+*		int nhosts
+*		int narches
+*		{
+*			int index
+*			string name
+*			string arch
+*			int mtu
+*			int speed
+*		} [nhosts]
+*	}
 */
 
 int
@@ -880,7 +1054,6 @@ tm_config(tp, mp)
 			pkint(mp, hp->hd_hostpart);
 			pkstr(mp, hp->hd_name);
 			pkstr(mp, hp->hd_arch ? hp->hd_arch : "");
-			pkint(mp, hp->hd_mtu);
 			pkint(mp, hp->hd_speed);
 		}
 	}
@@ -892,6 +1065,10 @@ tm_config(tp, mp)
 /*	tm_halt()
 *
 *	Command from task to stop master pvmd.
+*
+*	TM_HALT
+*	call { }
+*	ret { }
 */
 
 int
@@ -903,11 +1080,6 @@ tm_halt(tp, mp)
 	mp->m_cod = DM_HALT;
 	mp->m_dst = hosts->ht_hosts[hosts->ht_master]->hd_hostpart | TIDPVMD;
 	sendmessage(mp);
-
-	mp = mesg_new(0);
-	mp->m_cod = TM_HALT;
-	mp->m_dst = tp->t_tid;
-	sendmessage(mp);
 	return 0;
 }
 
@@ -915,6 +1087,21 @@ tm_halt(tp, mp)
 /*	tm_task()
 *
 *	Task wants a list of tasks.
+*	If where is a host or task tid, give a list of matching tasks.
+*	If where is zero, give all running tasks.
+*
+*	TM_TASK
+*	call {
+*		int where
+*	}
+*	ret {
+*		{
+*			int tid
+*			int ptid
+*			int hostpart
+*			int flag
+*		} [count]    // implied
+*	}
 */
 
 int
@@ -925,93 +1112,98 @@ tm_task(tp, mp)
 	int where;
 	struct task *tp2;
 	struct mesg *mp2;
-	struct waitc *wp, *wp2 = 0;
+	struct waitc *wp;
+	struct waitc *wp2 = 0;	/* master waitc of peer group */
 	int hh;
+	struct hostd *hp;
 
-	upkuint(mp, &where);
+	if (upkuint(mp, &where)) {
+		pvmlogerror("tm_task() bad msg format\n");
+		return 0;
+	}
 
 	mp = mesg_new(0);
 	mp->m_dst = tp->t_tid;
 	mp->m_cod = TM_TASK;
 
-	/* if where nonzero, find the host */
-
-	if (where && where != myhostpart) {
-		for (hh = hosts->ht_last; hh > 0; hh--)
-			if (hosts->ht_hosts[hh]
-			&& hosts->ht_hosts[hh]->hd_hostpart == where)
-				break;
-
-		if (hh < 1) {				/* err, no such host */
+	if (where) {		/* specific host or task requested */
+		if (!(hp = tidtohost(hosts, where))) {
 			pkint(mp, PvmNoHost);
 			sendmessage(mp);
 			return 0;
 		}
-
-	/* dial direct */
 		pkint(mp, 0);
 		wp = wait_new(WT_TASK);
 		wp->wa_mesg = mp;
 		wp->wa_tid = tp->t_tid;
-		wp->wa_on = hosts->ht_hosts[hh]->hd_hostpart;
+		wp->wa_on = hp->hd_hostpart;
 
 		mp = mesg_new(0);
 		mp->m_cod = DM_TASK;
-		mp->m_dst = hosts->ht_hosts[hh]->hd_hostpart | TIDPVMD;
+		mp->m_dst = hp->hd_hostpart | TIDPVMD;
 		mp->m_wid = wp->wa_wid;
+		pkint(mp, where);
 		sendmessage(mp);
-		return 0;
-	}
 
-	/* pack list of local tasks */
+	} else {			/* all tasks requested */
+		pkint(mp, 0);
 
-	pkint(mp, 0);	/* no error */
-	for (tp2 = locltasks->t_link; tp2 != locltasks; tp2 = tp2->t_link) {
-		pkint(mp, tp2->t_tid);
-		pkint(mp, tp2->t_ptid);
-		pkint(mp, myhostpart);
-		pkint(mp, tp2->t_flag);
-		pkstr(mp, tp2->t_a_out ? tp2->t_a_out : "");
-	}
-
-	/* if no other hosts, reply right away */
-
-	if (hosts->ht_cnt == 1 || where == myhostpart) {
-		sendmessage(mp);
-		return 0;
-	}
-
-	/* send request to each other host */
-
-	mp2 = mesg_new(0);
-	mp2->m_cod = DM_TASK;
-
-	for (hh = hosts->ht_last; hh > 0; hh--) {
-		if (hh == hosts->ht_local || !hosts->ht_hosts[hh])
-			continue;
-
-		wp = wait_new(WT_TASK);
+		wp2 = wait_new(WT_TASK);
 		mp->m_ref++;
-		wp->wa_mesg = mp;
-		wp->wa_tid = tp->t_tid;
-		wp->wa_on = hosts->ht_hosts[hh]->hd_hostpart;
+		wp2->wa_mesg = mp;
+		wp2->wa_tid = tp->t_tid;
+
+		mp2 = mesg_new(0);
+		mp2->m_cod = DM_TASK;
+		pkint(mp2, 0);
+
+		for (hh = hosts->ht_last; hh > 0; hh--) {
+			if (!hosts->ht_hosts[hh])
+				continue;
+
+			wp = wait_new(WT_TASK);
+			mp->m_ref++;
+			wp->wa_mesg = mp;
+			wp->wa_tid = tp->t_tid;
+			wp->wa_on = hosts->ht_hosts[hh]->hd_hostpart;
 
-		if (wp2) {
 			LISTPUTBEFORE(wp2, wp, wa_peer, wa_rpeer);
-		} else
-			wp2 = wp;
 
-		mp2->m_dst = hosts->ht_hosts[hh]->hd_hostpart | TIDPVMD;
-		mp2->m_wid = wp->wa_wid;
-		mp2->m_ref++;
-		sendmessage(mp2);
+			mp2->m_dst = hosts->ht_hosts[hh]->hd_hostpart | TIDPVMD;
+			mp2->m_wid = wp->wa_wid;
+			mp2->m_ref++;
+			sendmessage(mp2);
+		}
+		mesg_unref(mp2);
+		mesg_unref(mp);
+
+		/* send message if all waiters are in */
+
+		if (wp2->wa_peer == wp2) {
+			mp->m_ref++;
+			sendmessage(mp);
+		}
+		wait_delete(wp2);
 	}
-	mesg_unref(mp2);
-	mesg_unref(mp);
 	return 0;
 }
 
 
+/*	tm_tickle()
+*
+*	Task wants to poke at pvmd.
+*
+*	TM_TICKLE
+*	call {
+*		int nargs
+*		int args[nargs]
+*	}
+*	ret {
+*		int nresult
+*		int results[nresult]
+*	}
+*/
+
 int
 tm_tickle(tp, mp)
 	struct task *tp;
@@ -1021,6 +1213,7 @@ tm_tickle(tp, mp)
 	int arg[10];
 	int i;
 	char *p;
+	struct hostd *hp;
 
 	upkint(mp, &nar);
 	if (nar < 1 || nar > 10) {
@@ -1029,6 +1222,8 @@ tm_tickle(tp, mp)
 	}
 	for (i = 0; i < nar; i++)
 		upkint(mp, &arg[i]);
+	while (i < sizeof(arg)/sizeof(arg[0]))
+		arg[i++] = 0;
 
 	strcpy(pvmtxt, "tm_tickle() #");
 	p = pvmtxt + strlen(pvmtxt);
@@ -1081,6 +1276,45 @@ tm_tickle(tp, mp)
 		pkint(mp, 0);
 		break;
 
+	case 7:
+		if (arg[1] > 0 && arg[1] < 50) {
+			nopax = arg[1];
+			sprintf(pvmtxt, "tm_tickle() nopax is %d\n", nopax);
+		} else
+			sprintf(pvmtxt, "tm_tickle() bogus nopax %d\n", arg[1]);
+		pvmlogerror(pvmtxt);
+		pkint(mp, 0);
+		break;
+
+	case 8:
+		pkint(mp, 1);
+		if ((hp = tidtohost(hosts, arg[1]))
+		&& hp != hosts->ht_hosts[hosts->ht_local]) {
+			sprintf(pvmtxt, "tm_tickle() failing %s\n", hp->hd_name);
+			hostfailentry(hp);
+			ht_delete(hosts, hp);
+			pkint(mp, 1);
+		} else {
+			if (hp)
+				sprintf(pvmtxt, "tm_tickle() can't fail %s\n", hp->hd_name);
+			else
+				sprintf(pvmtxt, "tm_tickle() no such host %x\n", arg[1]);
+			pkint(mp, 0);
+		}
+		pvmlogerror(pvmtxt);
+		break;
+
+	case 9:
+#ifdef	STATISTICS
+		dump_statistics();
+		if (arg[1])
+			reset_statistics();
+#else
+		pvmlogerror("tm_tickle() statistics not compiled in\n");
+#endif
+		pkint(mp, 0);
+		break;
+
 	default:
 		sprintf(pvmtxt, "tm_tickle() don't know #%d\n", arg[0]);
 		pvmlogerror(pvmtxt);
@@ -1099,6 +1333,16 @@ tm_tickle(tp, mp)
 *
 *	Task requesting to add hosts to virtual machine.  Exit point is
 *	here or dm_delhostack().
+*
+*	TM_DELHOST
+*	call {
+*		int nhosts
+*		string names[nhosts]
+*	}
+*	ret {
+*		int nhosts			// or negative for error
+*		int status[nhosts]	// status of each host
+*	}
 */
 
 int
@@ -1150,6 +1394,20 @@ int_compare(i, j)
 }
 
 
+/*	tm_mca()
+*
+*	Task wants multicast address to use for subsequent message.
+*
+*	TM_MCA
+*	call {
+*		int count			// number of addresses
+*		int tids[count]		// addresses
+*	}
+*	ret {
+*		int gtid			// multicast tid to use
+*	}
+*/
+
 int
 tm_mca(tp, mp)
 	struct task *tp;
@@ -1170,25 +1428,28 @@ tm_mca(tp, mp)
 	/*
 	* unpack list of dst tids from message.
 	* normalize local dsts to nonzero hostpart.
-	* discard tids for nonexistent local tasks or foreign hosts.
+	* discard tids for nonexistent local tasks or foreign hosts or the sender.
 	*/
 
 	mcap = mca_new();
 	mcap->mc_tid = myhostpart | TIDGID | lastmca;
 	upkint(mp, &ndst);
 	dsts = TALLOC(ndst, int, "dsts");
-	for (i1 = 0; i1 < ndst; i1++) {
+	for (i1 = 0; i1 < ndst; ) {
 		upkuint(mp, &tid);
-		if (!(tid & tidhmask) || (tid & tidhmask) == myhostpart)
+		if (!(tid & tidhmask) || (tid & tidhmask) == myhostpart) {
 			tid |= myhostpart;
-
-		else
-			if (!tidtohost(hosts, tid)) {
-				i1--;
+			if (tid == tp->t_tid) {
 				ndst--;
 				continue;
 			}
-		dsts[i1] = tid;
+
+		} else
+			if (!tidtohost(hosts, tid)) {
+				ndst--;
+				continue;
+			}
+		dsts[i1++] = tid;
 	}
 
 	if (ndst < 1)		/* if no dsts left then just reply to task */
@@ -1196,6 +1457,7 @@ tm_mca(tp, mp)
 
 	/*
 	* sort dst tids
+	* XXX we could easily remove duplicates here
 	* send DM_MCA messages containing tids to dst hosts
 	* make list of dst hosts for us
 	*/
@@ -1243,6 +1505,20 @@ noguys:
 }
 
 
+/*	tm_notify()
+*
+*	Task wants to be notified when an event happens.
+*
+*	TM_NOTIFY
+*	call {
+*		int what			// event type
+*		int code			// message code to use in reply
+*		int count			// number of addresses
+*		int tids[count]		// addresses
+*	}
+*	// No Return
+*/
+
 int
 tm_notify(tp, mp)
 	struct task *tp;
@@ -1331,6 +1607,13 @@ dead:
 		}
 		break;
 
+	case PvmHostAdd:
+		wp = wait_new(WT_HOSTA);
+		wp->wa_tid = tp->t_tid;
+		wp->wa_on = tp->t_tid;
+		wp->wa_dep = code;
+		break;
+
 	default:
 		sprintf(pvmtxt, "tm_notify() unknown what=%d\n", what);
 		pvmlogerror(pvmtxt);
@@ -1340,6 +1623,19 @@ dead:
 }
 
 
+/*	tm_mstat()
+*
+*	Task wants status of a host.
+*
+*	TM_MSTAT
+*	call {
+*		string hostname
+*	}
+*	ret {
+*		int status
+*	}
+*/
+
 int
 tm_mstat(tp, mp)
 	struct task *tp;
@@ -1347,7 +1643,6 @@ tm_mstat(tp, mp)
 {
 	char *name;
 	struct hostd *hp;
-	int hh;
 	struct waitc *wp;
 
 	if (upkstralloc(mp, &name)) {
@@ -1355,13 +1650,11 @@ tm_mstat(tp, mp)
 		return 0;
 	}
 
-	for (hh = hosts->ht_last; hh > 0; hh--)
-		if ((hp = hosts->ht_hosts[hh]) && !strcmp(name, hp->hd_name))
-			break;
+	hp = nametohost(hosts, name);
 
 	PVM_FREE(name);
 
-	if (hh < 1) {
+	if (!hp) {
 		mp = mesg_new(0);
 		mp->m_dst = tp->t_tid;
 		mp->m_cod = TM_MSTAT;
@@ -1384,6 +1677,23 @@ tm_mstat(tp, mp)
 }
 
 
+/*	tm_db()
+*
+*	Task wants a database op.
+*
+*	TM_DB
+*	call {
+*		int opcode		// insert, delete, lookup
+*		string name
+*		int index
+*		int data		// if 'insert'
+*	}
+*	ret {
+*		int index		// and status
+*		int data		// if 'lookup'
+*	}
+*/
+
 int
 tm_db(tp, mp)
 	struct task *tp;
